

class ToolCache(object):
    """
    Cache tool definitions to allow quickly reloading the whole
    toolbox.
    """

    def __init__(self):
        self._tools_by_path = {}
        self._tool_paths_by_id = {}

    def get_tool(self, config_filename):
        """ Get the tool from the cache if the tool is up to date.
        """
        return self._tools_by_path.get(config_filename, None)

    def expire_tool(self, tool_id):
        if tool_id in self._tool_paths_by_id:
            config_filename = self._tool_paths_by_id[tool_id]
            del self._tool_paths_by_id[tool_id]
            del self._tools_by_path[config_filename]

    def cache_tool(self, config_filename, tool):
        tool_id = str( tool.id )
        self._tool_paths_by_id[tool_id] = config_filename
        self._tools_by_path[config_filename] = tool



import os
import tempfile

from abc import (
    ABCMeta,
    abstractmethod,
    abstractproperty,
)

import six


@six.add_metaclass(ABCMeta)
class ToolLocationResolver(object):
    """Parse a URI-like string and return a ToolSource object."""

    @abstractproperty
    def scheme(self):
        """Short label for the type of location resolver and URI scheme."""

    @abstractmethod
    def get_tool_source_path(self, uri_like):
        """Return a local path for the uri_like string."""

    def _temp_path(self, uri_like):
        """Create an abstraction for this so we can configure and cache later."""
        handle, filename = tempfile.mkstemp(suffix=uri_like.split("/")[-1])
        os.close(handle)
        return filename



"""This module contains a linting functions for tool inputs."""
from galaxy.util import string_as_bool

from ..lint_util import is_datasource


def lint_inputs(tool_xml, lint_ctx):
    """Lint parameters in a tool's inputs block."""
    datasource = is_datasource(tool_xml)
    inputs = tool_xml.findall("./inputs//param")
    num_inputs = 0
    for param in inputs:
        num_inputs += 1
        param_attrib = param.attrib
        has_errors = False
        if "type" not in param_attrib:
            lint_ctx.error("Found param input with no type specified.")
            has_errors = True
        if "name" not in param_attrib and "argument" not in param_attrib:
            lint_ctx.error("Found param input with no name specified.")
            has_errors = True

        if has_errors:
            continue

        param_type = param_attrib["type"]
        param_name = param_attrib.get("name", param_attrib.get("argument"))
        if param_type == "data":
            if "format" not in param_attrib:
                lint_ctx.warn("Param input [%s] with no format specified - 'data' format will be assumed.", param_name)

        if param_type == "select":
            dynamic_options = param.get("dynamic_options", None)
            if dynamic_options is None:
                dynamic_options = param.find("options")

            select_options = _find_with_attribute(param, 'option', 'value')
            if any(['value' not in option.attrib for option in select_options]):
                lint_ctx.error("Option without value")

            if dynamic_options is None and len(select_options) == 0:
                message = "No options defined for select [%s]" % param_name
                lint_ctx.warn(message)

        # TODO: Validate type, much more...

    conditional_selects = tool_xml.findall("./inputs//conditional")
    for conditional in conditional_selects:
        conditional_name = conditional.get('name')
        if not conditional_name:
            lint_ctx.error("Conditional without a name")
        if conditional.get("value_from"):
            # Probably only the upload tool use this, no children elements
            continue
        first_param = conditional.find("param")
        if first_param is None:
            lint_ctx.error("Conditional '%s' has no child <param>" % conditional_name)
            continue
        first_param_type = first_param.get('type')
        if first_param_type not in ['select', 'boolean']:
            lint_ctx.warn("Conditional '%s' first param should have type=\"select\" /> or type=\"boolean\"" % conditional_name)
            continue

        if first_param_type == 'select':
            select_options = _find_with_attribute(first_param, 'option', 'value')
            option_ids = [option.get('value') for option in select_options]
        else:  # boolean
            option_ids = [
                first_param.get('truevalue', 'true'),
                first_param.get('falsevalue', 'false')
            ]

        if string_as_bool(first_param.get('optional', False)):
            lint_ctx.warn("Conditional test parameter cannot be optional")

        whens = conditional.findall('./when')
        if any('value' not in when.attrib for when in whens):
            lint_ctx.error("When without value")

        when_ids = [w.get('value') for w in whens]

        for option_id in option_ids:
            if option_id not in when_ids:
                lint_ctx.warn("No <when /> block found for %s option '%s' inside conditional '%s'" % (first_param_type, option_id, conditional_name))

        for when_id in when_ids:
            if when_id not in option_ids:
                if first_param_type == 'select':
                    lint_ctx.warn("No <option /> found for when block '%s' inside conditional '%s'" % (when_id, conditional_name))
                else:
                    lint_ctx.warn("No truevalue/falsevalue found for when block '%s' inside conditional '%s'" % (when_id, conditional_name))

    if datasource:
        for datasource_tag in ('display', 'uihints'):
            if not any([param.tag == datasource_tag for param in inputs]):
                lint_ctx.info("%s tag usually present in data sources" % datasource_tag)

    if num_inputs:
        lint_ctx.info("Found %d input parameters.", num_inputs)
    else:
        if datasource:
            lint_ctx.info("No input parameters, OK for data sources")
        else:
            lint_ctx.warn("Found no input parameters.")


def lint_repeats(tool_xml, lint_ctx):
    """Lint repeat blocks in tool inputs."""
    repeats = tool_xml.findall("./inputs//repeat")
    for repeat in repeats:
        if "name" not in repeat.attrib:
            lint_ctx.error("Repeat does not specify name attribute.")
        if "title" not in repeat.attrib:
            lint_ctx.error("Repeat does not specify title attribute.")


def _find_with_attribute(element, tag, attribute, test_value=None):
    rval = []
    for el in (element.findall('./%s' % tag) or []):
        if attribute not in el.attrib:
            continue
        value = el.attrib[attribute]
        if test_value is not None:
            if value == test_value:
                rval.append(el)
        else:
            rval.append(el)
    return rval



"""This module describes the :class:`MulledContainerResolver` ContainerResolver plugin."""

import collections
import logging

from ..container_resolvers import (
    ContainerResolver,
)
from ..docker_util import build_docker_images_command
from ..mulled.mulled_build import (
    check_output,
    DEFAULT_CHANNELS,
    ensure_installed,
    InvolucroContext,
    mull_targets,
)
from ..mulled.mulled_build_tool import requirements_to_mulled_targets
from ..mulled.util import (
    image_name,
    mulled_tags_for,
    split_tag,
)
from ..requirements import ContainerDescription

log = logging.getLogger(__name__)


CachedMulledImageSingleTarget = collections.namedtuple("CachedMulledImageSingleTarget", ["package_name", "version", "build", "image_identifier"])
CachedMulledImageMultiTarget = collections.namedtuple("CachedMulledImageMultiTarget", ["hash", "image_identifier"])

CachedMulledImageSingleTarget.multi_target = False
CachedMulledImageMultiTarget.multi_target = True


def list_cached_mulled_images(namespace=None):
    command = build_docker_images_command(truncate=True, sudo=False)
    command = "%s | tail -n +2 | tr -s ' ' | cut -d' ' -f1,2" % command
    images_and_versions = check_output(command)
    name_filter = get_filter(namespace)

    def output_line_to_image(line):
        image_name, version = line.split(" ", 1)
        identifier = "%s:%s" % (image_name, version)
        url, namespace, package_description = image_name.split("/")

        if package_description.startswith("mulled-v1-"):
            hash = package_description
            image = CachedMulledImageMultiTarget(hash, identifier)
        else:
            build = None
            if not version or version == "latest":
                version = None

            if version and "--" in version:
                version, build = split_tag(version)

            image = CachedMulledImageSingleTarget(image_name, version, build, identifier)

        return image

    return [output_line_to_image(_) for _ in filter(name_filter, images_and_versions.splitlines())]


def get_filter(namespace):
    prefix = "quay.io/" if namespace is None else "quay.io/%s" % namespace
    return lambda name: name.startswith(prefix) and name.count("/") == 2


def cached_container_description(targets, namespace):
    if len(targets) == 0:
        return None

    cached_images = list_cached_mulled_images(namespace)
    image = None
    if len(targets) == 1:
        target = targets[0]
        for cached_image in cached_images:
            if cached_image.multi_target:
                continue
            if not cached_image.package_name == target.package_name:
                continue
            if not target.version or target.version == cached_image.version:
                image = cached_image
                break
    else:
        name = image_name(targets)
        for cached_image in cached_images:
            if not cached_image.multi_target:
                continue

            if name == cached_image.hash:
                image = cached_image
                break

    container = None
    if image:
        container = ContainerDescription(
            image.image_identifier,
            type="docker",
        )

    return container


class CachedMulledContainerResolver(ContainerResolver):

    resolver_type = "cached_mulled"

    def __init__(self, app_info=None, namespace=None):
        super(CachedMulledContainerResolver, self).__init__(app_info)
        self.namespace = namespace

    def resolve(self, enabled_container_types, tool_info):
        targets = mulled_targets(tool_info)
        return cached_container_description(targets, self.namespace)


class MulledContainerResolver(ContainerResolver):
    """Look for mulled images matching tool dependencies."""

    resolver_type = "mulled"

    def __init__(self, app_info=None, namespace="biocontainers"):
        super(MulledContainerResolver, self).__init__(app_info)
        self.namespace = namespace

    def resolve(self, enabled_container_types, tool_info):
        targets = mulled_targets(tool_info)
        if len(targets) == 0:
            return None

        name = None

        if len(targets) == 1:
            target = targets[0]
            target_version = target.version
            tags = mulled_tags_for(self.namespace, target.package_name)

            if not tags:
                return None

            if target_version:
                for tag in tags:
                    version, build = split_tag(tag)
                    if version == target_version:
                        name = "%s:%s--%s" % (target.package_name, version, build)
                        break
            else:
                version, build = split_tag(tags[0])
                name = "%s:%s--%s" % (target.package_name, version, build)
        else:
            base_image_name = image_name(targets)
            tags = mulled_tags_for(self.namespace, base_image_name)
            if tags:
                name = "%s:%s" % (base_image_name, tags[0])

        if name:
            return ContainerDescription(
                "quay.io/%s/%s" % (self.namespace, name),
                type="docker",
            )


class BuildMulledContainerResolver(ContainerResolver):
    """Look for mulled images matching tool dependencies."""

    resolver_type = "build_mulled"

    def __init__(self, app_info=None, namespace="local", **kwds):
        super(BuildMulledContainerResolver, self).__init__(app_info)
        self._involucro_context_kwds = {
            'involucro_bin': self._get_config_option("involucro_path", None)
        }
        self.namespace = namespace
        self._mulled_kwds = {
            'namespace': namespace,
            'channels': self._get_config_option("channels", DEFAULT_CHANNELS, prefix="mulled"),
        }
        self.auto_init = self._get_config_option("auto_init", DEFAULT_CHANNELS, prefix="involucro")

    def resolve(self, enabled_container_types, tool_info):
        targets = mulled_targets(tool_info)
        if len(targets) == 0:
            return None

        mull_targets(
            targets,
            involucro_context=self._get_involucro_context(),
            **self._mulled_kwds
        )
        return cached_container_description(targets, self.namespace)

    def _get_involucro_context(self):
        involucro_context = InvolucroContext(**self._involucro_context_kwds)
        self.enabled = ensure_installed(involucro_context, self.auto_init)
        return involucro_context


def mulled_targets(tool_info):
    return requirements_to_mulled_targets(tool_info.requirements)


__all__ = (
    "CachedMulledContainerResolver",
    "MulledContainerResolver",
    "BuildMulledContainerResolver",
)



import os

from ..brew_exts import (
    build_env_statements,
    DEFAULT_HOMEBREW_ROOT,
    recipe_cellar_path,
)
from ..resolvers import Dependency, NullDependency


class UsesHomebrewMixin:

    def _init_homebrew(self, **kwds):
        cellar_root = kwds.get('cellar', None)
        if cellar_root is None:
            cellar_root = os.path.join(DEFAULT_HOMEBREW_ROOT, "Cellar")

        self.cellar_root = cellar_root

    def _find_dep_versioned(self, name, version):
        recipe_path = recipe_cellar_path(self.cellar_root, name, version)
        if not os.path.exists(recipe_path) or not os.path.isdir(recipe_path):
            return NullDependency(version=version, name=name)

        commands = build_env_statements(self.cellar_root, recipe_path, relaxed=True)
        return HomebrewDependency(commands)

    def _find_dep_default(self, name, version):
        installed_versions = self._installed_versions(name)
        if not installed_versions:
            return NullDependency(version=version, name=name)

        # Just grab newest installed version - may make sense some day to find
        # the linked version instead.
        default_version = sorted(installed_versions, reverse=True)[0]
        return self._find_dep_versioned(name, default_version, exact=version is None)

    def _installed_versions(self, recipe):
        recipe_base_path = os.path.join(self.cellar_root, recipe)
        if not os.path.exists(recipe_base_path):
            return []

        names = os.listdir(recipe_base_path)
        return [n for n in names if os.path.isdir(os.path.join(recipe_base_path, n))]


class UsesToolDependencyDirMixin:

    def _init_base_path(self, dependency_manager, **kwds):
        self.base_path = os.path.abspath( kwds.get('base_path', dependency_manager.default_base_path) )


class UsesInstalledRepositoriesMixin:

    def _get_installed_dependency( self, name, type, version=None, **kwds ):
        installed_tool_dependencies = kwds.get("installed_tool_dependencies") or []
        for installed_tool_dependency in installed_tool_dependencies:
            if installed_tool_dependency.name == name and installed_tool_dependency.type == type:
                if not version or installed_tool_dependency.version == version:
                    return installed_tool_dependency
        return None


class HomebrewDependency(Dependency):

    def __init__(self, commands, exact=True):
        self.commands = commands
        self._exact = exact

    @property
    def exact(self):
        return self._exact

    def shell_commands(self, requirement):
        raw_commands = self.commands.replace("\n", ";")
        return raw_commands

    def __repr__(self):
        return "PlatformBrewDependency[commands=%s]" % self.commands



from .interface import ToolStdioExitCode
from .interface import ToolStdioRegex


def error_on_exit_code():
    exit_code_lower = ToolStdioExitCode()
    exit_code_lower.range_start = float("-inf")
    exit_code_lower.range_end = -1
    _set_fatal(exit_code_lower)
    exit_code_high = ToolStdioExitCode()
    exit_code_high.range_start = 1
    exit_code_high.range_end = float("inf")
    _set_fatal(exit_code_high)
    return [exit_code_lower, exit_code_high], []


def aggressive_error_checks():
    exit_codes, _ = error_on_exit_code()
    # these regexes are processed as case insensitive by default
    regexes = [
        _error_regex("exception:"),
        _error_regex("error:")
    ]
    return exit_codes, regexes


def _error_regex(match):
    regex = ToolStdioRegex()
    _set_fatal(regex)
    regex.match = match
    regex.stdout_match = True
    regex.stderr_match = True
    return regex


def _set_fatal(obj):
    from galaxy.jobs.error_level import StdioErrorLevel
    obj.error_level = StdioErrorLevel.FATAL



"""The module describes the ``meminfo`` job metrics plugin."""
import re
import sys

from galaxy import util

from ..instrumenters import InstrumentPlugin
from ...metrics import formatting

if sys.version_info > (3,):
    long = int


MEMINFO_LINE = re.compile(r"(\w+)\s*\:\s*(\d+) kB")

# Important (non-verbose) meminfo property titles.
MEMINFO_TITLES = {
    "memtotal": "Total System Memory",
    "swaptotal": "Total System Swap"
}


class MemInfoFormatter( formatting.JobMetricFormatter ):

    def format( self, key, value ):
        title = MEMINFO_TITLES.get( key, key )
        return title, util.nice_size( value * 1000 )  # kB = *1000, KB = *1024 - wikipedia


class MemInfoPlugin( InstrumentPlugin ):
    """ Gather information about processor configuration from /proc/cpuinfo.
    Linux only.
    """
    plugin_type = "meminfo"
    formatter = MemInfoFormatter()

    def __init__( self, **kwargs ):
        self.verbose = util.asbool( kwargs.get( "verbose", False ) )

    def pre_execute_instrument( self, job_directory ):
        return "cat /proc/meminfo > '%s'" % self.__instrument_meminfo_path( job_directory )

    def job_properties( self, job_id, job_directory ):
        properties = {}
        with open( self.__instrument_meminfo_path( job_directory ) ) as f:
            for line in f:
                line = line.strip()
                if not line:  # Skip empty lines
                    continue
                line_match = MEMINFO_LINE.match( line )
                if not line_match:
                    continue
                key = line_match.group( 1 ).lower()
                # By default just grab important meminfo properties with titles
                # defined for formatter. Grab everything in verbose mode for
                # an arbitrary snapshot of memory at beginning of run.
                if key in MEMINFO_TITLES or self.verbose:
                    value = long( line_match.group( 2 ) )
                    properties[ key ] = value
        return properties

    def __instrument_meminfo_path( self, job_directory ):
        return self._instrument_file_path( job_directory, "meminfo" )


__all__ = ( 'MemInfoPlugin', )



import threading


class Sleeper( object ):
    """
    Provides a 'sleep' method that sleeps for a number of seconds *unless*
    the notify method is called (from a different thread).
    """
    def __init__( self ):
        self.condition = threading.Condition()

    def sleep( self, seconds ):
        # Should this be in a try/finally block? -John
        self.condition.acquire()
        self.condition.wait( seconds )
        self.condition.release()

    def wake( self ):
        # Should this be in a try/finally block? -John
        self.condition.acquire()
        self.condition.notify()
        self.condition.release()



import logging
import re
import sys
import traceback
import uuid

from math import isinf

from galaxy.tools.deps import requirements
from galaxy.util import string_as_bool, xml_text, xml_to_string
from galaxy.util.odict import odict

from .interface import (
    InputSource,
    PageSource,
    PagesSource,
    TestCollectionDef,
    TestCollectionOutputDef,
    ToolSource,
    ToolStdioExitCode,
    ToolStdioRegex,
)
from .output_actions import ToolOutputActionGroup
from .output_collection_def import dataset_collector_descriptions_from_elem
from .output_objects import (
    ToolOutput,
    ToolOutputCollection,
    ToolOutputCollectionStructure
)
from .util import (
    aggressive_error_checks,
    error_on_exit_code,
)


log = logging.getLogger( __name__ )


class XmlToolSource(ToolSource):
    """ Responsible for parsing a tool from classic Galaxy representation.
    """

    def __init__(self, xml_tree, source_path=None):
        self.xml_tree = xml_tree
        self.root = xml_tree.getroot()
        self._source_path = source_path
        self.legacy_defaults = self.parse_profile() == "16.01"

    def parse_version(self):
        return self.root.get("version", None)

    def parse_id(self):
        return self.root.get("id")

    def parse_tool_module(self):
        root = self.root
        if root.find( "type" ) is not None:
            type_elem = root.find( "type" )
            module = type_elem.get( 'module', 'galaxy.tools' )
            cls = type_elem.get( 'class' )
            return module, cls

        return None

    def parse_action_module(self):
        root = self.root
        action_elem = root.find( "action" )
        if action_elem is not None:
            module = action_elem.get( 'module' )
            cls = action_elem.get( 'class' )
            return module, cls
        else:
            return None

    def parse_tool_type(self):
        root = self.root
        if root.get( 'tool_type', None ) is not None:
            return root.get( 'tool_type' )

    def parse_name(self):
        return self.root.get( "name" )

    def parse_edam_operations(self):
        edam_ops = self.root.find("edam_operations")
        if edam_ops is None:
            return []
        return [ edam_op.text for edam_op in edam_ops.findall("edam_operation") ]

    def parse_edam_topics(self):
        edam_topics = self.root.find("edam_topics")
        if edam_topics is None:
            return []
        return [ edam_topic.text for edam_topic in edam_topics.findall("edam_topic") ]

    def parse_description(self):
        return xml_text(self.root, "description")

    def parse_is_multi_byte(self):
        return self._get_attribute_as_bool( "is_multi_byte", self.default_is_multi_byte )

    def parse_display_interface(self, default):
        return self._get_attribute_as_bool( "display_interface", default )

    def parse_require_login(self, default):
        return self._get_attribute_as_bool( "require_login", default )

    def parse_request_param_translation_elem(self):
        return self.root.find( "request_param_translation" )

    def parse_command(self):
        command_el = self._command_el
        return ( ( command_el is not None ) and command_el.text ) or None

    def parse_environment_variables(self):
        environment_variables_el = self.root.find("environment_variables")
        if environment_variables_el is None:
            return []

        environment_variables = []
        for environment_variable_el in environment_variables_el.findall("environment_variable"):
            definition = {
                "name": environment_variable_el.get("name"),
                "template": environment_variable_el.text,
            }
            environment_variables.append(
                definition
            )
        return environment_variables

    def parse_interpreter(self):
        command_el = self._command_el
        interpreter = (command_el is not None) and command_el.get("interpreter", None)
        if not self.legacy_defaults:
            log.warning("Deprecated interpeter attribute on command element is now ignored.")
            interpreter = None

        return interpreter

    def parse_version_command(self):
        version_cmd = self.root.find("version_command")
        if version_cmd is not None:
            return version_cmd.text
        else:
            return None

    def parse_version_command_interpreter(self):
        if self.parse_version_command() is not None:
            version_cmd = self.root.find("version_command")
            version_cmd_interpreter = version_cmd.get( "interpreter", None )
            if version_cmd_interpreter:
                return version_cmd_interpreter
        return None

    def parse_parallelism(self):
        parallelism = self.root.find("parallelism")
        parallelism_info = None
        if parallelism is not None and parallelism.get("method"):
            from galaxy.jobs import ParallelismInfo
            return ParallelismInfo(parallelism)
        return parallelism_info

    def parse_hidden(self):
        hidden = xml_text(self.root, "hidden")
        if hidden:
            hidden = string_as_bool(hidden)
        return hidden

    def parse_redirect_url_params_elem(self):
        return self.root.find("redirect_url_params")

    def parse_sanitize(self):
        return self._get_option_value("sanitize", True)

    def parse_refresh(self):
        return self._get_option_value("refresh", False)

    def _get_option_value(self, key, default):
        root = self.root
        for option_elem in root.findall("options"):
            if key in option_elem.attrib:
                return string_as_bool(option_elem.get(key))

        return default

    @property
    def _command_el(self):
        return self.root.find("command")

    def _get_attribute_as_bool( self, attribute, default, elem=None ):
        if elem is None:
            elem = self.root
        return string_as_bool( elem.get( attribute, default ) )

    def parse_requirements_and_containers(self):
        return requirements.parse_requirements_from_xml(self.root)

    def parse_input_pages(self):
        return XmlPagesSource(self.root)

    def parse_outputs(self, tool):
        out_elem = self.root.find("outputs")
        outputs = odict()
        output_collections = odict()
        if out_elem is None:
            return outputs, output_collections

        data_dict = odict()

        def _parse(data_elem, **kwds):
            output_def = self._parse_output(data_elem, tool, **kwds)
            data_dict[output_def.name] = output_def
            return output_def

        map(_parse, out_elem.findall("data"))

        for collection_elem in out_elem.findall("collection"):
            name = collection_elem.get( "name" )
            label = xml_text( collection_elem, "label" )
            default_format = collection_elem.get( "format", "data" )
            collection_type = collection_elem.get( "type", None )
            collection_type_source = collection_elem.get( "type_source", None )
            structured_like = collection_elem.get( "structured_like", None )
            inherit_format = False
            inherit_metadata = False
            if structured_like:
                inherit_format = string_as_bool( collection_elem.get( "inherit_format", None ) )
                inherit_metadata = string_as_bool( collection_elem.get( "inherit_metadata", None ) )
            default_format_source = collection_elem.get( "format_source", None )
            default_metadata_source = collection_elem.get( "metadata_source", "" )
            filters = collection_elem.findall( 'filter' )

            dataset_collector_descriptions = None
            if collection_elem.find( "discover_datasets" ) is not None:
                dataset_collector_descriptions = dataset_collector_descriptions_from_elem( collection_elem, legacy=False )
            structure = ToolOutputCollectionStructure(
                collection_type=collection_type,
                collection_type_source=collection_type_source,
                structured_like=structured_like,
                dataset_collector_descriptions=dataset_collector_descriptions,
            )
            output_collection = ToolOutputCollection(
                name,
                structure,
                label=label,
                filters=filters,
                default_format=default_format,
                inherit_format=inherit_format,
                inherit_metadata=inherit_metadata,
                default_format_source=default_format_source,
                default_metadata_source=default_metadata_source,
            )
            outputs[output_collection.name] = output_collection

            for data_elem in collection_elem.findall("data"):
                _parse(
                    data_elem,
                    default_format=default_format,
                    default_format_source=default_format_source,
                    default_metadata_source=default_metadata_source,
                )

            for data_elem in collection_elem.findall("data"):
                output_name = data_elem.get("name")
                data = data_dict[output_name]
                assert data
                del data_dict[output_name]
                output_collection.outputs[output_name] = data
            output_collections[ name ] = output_collection

        for output_def in data_dict.values():
            outputs[output_def.name] = output_def
        return outputs, output_collections

    def _parse_output(
        self,
        data_elem,
        tool,
        default_format="data",
        default_format_source=None,
        default_metadata_source="",
    ):
        output = ToolOutput( data_elem.get("name") )
        output_format = data_elem.get("format", default_format)
        auto_format = string_as_bool( data_elem.get( "auto_format", "false" ) )
        if auto_format and output_format != "data":
            raise ValueError("Setting format and auto_format is not supported at this time.")
        elif auto_format:
            output_format = "_sniff_"
        output.format = output_format
        output.change_format = data_elem.findall("change_format")
        output.format_source = data_elem.get("format_source", default_format_source)
        output.metadata_source = data_elem.get("metadata_source", default_metadata_source)
        output.parent = data_elem.get("parent", None)
        output.label = xml_text( data_elem, "label" )
        output.count = int( data_elem.get("count", 1) )
        output.filters = data_elem.findall( 'filter' )
        output.tool = tool
        output.from_work_dir = data_elem.get("from_work_dir", None)
        output.hidden = string_as_bool( data_elem.get("hidden", "") )
        output.actions = ToolOutputActionGroup( output, data_elem.find( 'actions' ) )
        output.dataset_collector_descriptions = dataset_collector_descriptions_from_elem( data_elem, legacy=self.legacy_defaults )
        return output

    def parse_stdio(self):
        command_el = self._command_el
        detect_errors = None
        if command_el is not None:
            detect_errors = command_el.get("detect_errors")
        if detect_errors and detect_errors != "default":
            if detect_errors == "exit_code":
                return error_on_exit_code()
            elif detect_errors == "aggressive":
                return aggressive_error_checks()
            else:
                raise ValueError("Unknown detect_errors value encountered [%s]" % detect_errors)
        elif len(self.root.findall('stdio')) == 0 and not self.legacy_defaults:
            return error_on_exit_code()
        else:
            parser = StdioParser(self.root)
            return parser.stdio_exit_codes, parser.stdio_regexes

    def parse_strict_shell(self):
        command_el = self._command_el
        if command_el is not None:
            return string_as_bool(command_el.get("strict", "False"))
        elif self.legacy_defaults:
            return False
        else:
            return True

    def parse_help(self):
        help_elem = self.root.find( 'help' )
        return help_elem.text if help_elem is not None else None

    def parse_tests_to_dict(self):
        tests_elem = self.root.find("tests")
        tests = []
        rval = dict(
            tests=tests
        )

        if tests_elem is not None:
            for i, test_elem in enumerate(tests_elem.findall("test")):
                tests.append(_test_elem_to_dict(test_elem, i))

            _copy_to_dict_if_present(tests_elem, rval, ["interactor"])

        return rval

    def parse_profile(self):
        # Pre-16.04 or default XML defaults
        # - Use standard error for error detection.
        # - Don't run shells with -e
        # - Auto-check for implicit multiple outputs.
        # - Auto-check for $param_file.
        # - Enable buggy interpreter attribute.
        return self.root.get("profile", "16.01")


def _test_elem_to_dict(test_elem, i):
    rval = dict(
        outputs=__parse_output_elems(test_elem),
        output_collections=__parse_output_collection_elems(test_elem),
        inputs=__parse_input_elems(test_elem, i),
        expect_num_outputs=test_elem.get("expect_num_outputs"),
        command=__parse_assert_list_from_elem( test_elem.find("assert_command") ),
        stdout=__parse_assert_list_from_elem( test_elem.find("assert_stdout") ),
        stderr=__parse_assert_list_from_elem( test_elem.find("assert_stderr") ),
        expect_exit_code=test_elem.get("expect_exit_code"),
        expect_failure=string_as_bool(test_elem.get("expect_failure", False)),
        maxseconds=test_elem.get("maxseconds", None),
    )
    _copy_to_dict_if_present(test_elem, rval, ["interactor", "num_outputs"])
    return rval


def __parse_input_elems(test_elem, i):
    __expand_input_elems( test_elem )
    return __parse_inputs_elems( test_elem, i )


def __parse_output_elems( test_elem ):
    outputs = []
    for output_elem in test_elem.findall( "output" ):
        name, file, attributes = __parse_output_elem( output_elem )
        outputs.append( ( name, file, attributes ) )
    return outputs


def __parse_output_elem( output_elem ):
    attrib = dict( output_elem.attrib )
    name = attrib.pop( 'name', None )
    if name is None:
        raise Exception( "Test output does not have a 'name'" )

    file, attributes = __parse_test_attributes( output_elem, attrib, parse_discovered_datasets=True )
    return name, file, attributes


def __parse_command_elem( test_elem ):
    assert_elem = test_elem.find("command")
    return __parse_assert_list_from_elem( assert_elem )


def __parse_output_collection_elems( test_elem ):
    output_collections = []
    for output_collection_elem in test_elem.findall( "output_collection" ):
        output_collection_def = __parse_output_collection_elem( output_collection_elem )
        output_collections.append( output_collection_def )
    return output_collections


def __parse_output_collection_elem( output_collection_elem ):
    attrib = dict( output_collection_elem.attrib )
    name = attrib.pop( 'name', None )
    if name is None:
        raise Exception( "Test output collection does not have a 'name'" )
    element_tests = __parse_element_tests( output_collection_elem )
    return TestCollectionOutputDef( name, attrib, element_tests )


def __parse_element_tests( parent_element ):
    element_tests = {}
    for element in parent_element.findall("element"):
        element_attrib = dict( element.attrib )
        identifier = element_attrib.pop( 'name', None )
        if identifier is None:
            raise Exception( "Test primary dataset does not have a 'identifier'" )
        element_tests[ identifier ] = __parse_test_attributes( element, element_attrib, parse_elements=True )
    return element_tests


def __parse_test_attributes( output_elem, attrib, parse_elements=False, parse_discovered_datasets=False ):
    assert_list = __parse_assert_list( output_elem )

    # Allow either file or value to specify a target file to compare result with
    # file was traditionally used by outputs and value by extra files.
    file = attrib.pop( 'file', attrib.pop( 'value', None ) )

    # File no longer required if an list of assertions was present.
    attributes = {}
    # Method of comparison
    attributes['compare'] = attrib.pop( 'compare', 'diff' ).lower()
    # Number of lines to allow to vary in logs (for dates, etc)
    attributes['lines_diff'] = int( attrib.pop( 'lines_diff', '0' ) )
    # Allow a file size to vary if sim_size compare
    attributes['delta'] = int( attrib.pop( 'delta', '10000' ) )
    attributes['sort'] = string_as_bool( attrib.pop( 'sort', False ) )
    attributes['decompress'] = string_as_bool( attrib.pop( 'decompress', False ) )
    extra_files = []
    if 'ftype' in attrib:
        attributes['ftype'] = attrib['ftype']
    for extra in output_elem.findall( 'extra_files' ):
        extra_files.append( __parse_extra_files_elem( extra ) )
    metadata = {}
    for metadata_elem in output_elem.findall( 'metadata' ):
        metadata[ metadata_elem.get('name') ] = metadata_elem.get( 'value' )
    md5sum = attrib.get("md5", None)
    checksum = attrib.get("checksum", None)
    element_tests = {}
    if parse_elements:
        element_tests = __parse_element_tests( output_elem )

    primary_datasets = {}
    if parse_discovered_datasets:
        for primary_elem in ( output_elem.findall( "discovered_dataset" ) or [] ):
            primary_attrib = dict( primary_elem.attrib )
            designation = primary_attrib.pop( 'designation', None )
            if designation is None:
                raise Exception( "Test primary dataset does not have a 'designation'" )
            primary_datasets[ designation ] = __parse_test_attributes( primary_elem, primary_attrib )

    has_checksum = md5sum or checksum
    has_nested_tests = extra_files or element_tests or primary_datasets
    if not (assert_list or file or metadata or has_checksum or has_nested_tests):
        raise Exception( "Test output defines nothing to check (e.g. must have a 'file' check against, assertions to check, metadata or checksum tests, etc...)")
    attributes['assert_list'] = assert_list
    attributes['extra_files'] = extra_files
    attributes['metadata'] = metadata
    attributes['md5'] = md5sum
    attributes['checksum'] = checksum
    attributes['elements'] = element_tests
    attributes['primary_datasets'] = primary_datasets
    return file, attributes


def __parse_assert_list( output_elem ):
    assert_elem = output_elem.find("assert_contents")
    return __parse_assert_list_from_elem( assert_elem )


def __parse_assert_list_from_elem( assert_elem ):
    assert_list = None

    def convert_elem(elem):
        """ Converts and XML element to a dictionary format, used by assertion checking code. """
        tag = elem.tag
        attributes = dict( elem.attrib )
        child_elems = list( elem.getchildren() )
        converted_children = []
        for child_elem in child_elems:
            converted_children.append( convert_elem(child_elem) )
        return {"tag": tag, "attributes": attributes, "children": converted_children}
    if assert_elem is not None:
        assert_list = []
        for assert_child in list(assert_elem):
            assert_list.append(convert_elem(assert_child))

    return assert_list


def __parse_extra_files_elem(extra):
    # File or directory, when directory, compare basename
    # by basename
    attrib = dict(extra.attrib)
    extra_type = attrib.pop('type', 'file')
    extra_name = attrib.pop('name', None)
    assert extra_type == 'directory' or extra_name is not None, \
        'extra_files type (%s) requires a name attribute' % extra_type
    extra_value, extra_attributes = __parse_test_attributes(extra, attrib)
    return extra_type, extra_value, extra_name, extra_attributes


def __expand_input_elems( root_elem, prefix="" ):
    __append_prefix_to_params( root_elem, prefix )

    repeat_elems = root_elem.findall( 'repeat' )
    indices = {}
    for repeat_elem in repeat_elems:
        name = repeat_elem.get( "name" )
        if name not in indices:
            indices[ name ] = 0
            index = 0
        else:
            index = indices[ name ] + 1
            indices[ name ] = index

        new_prefix = __prefix_join( prefix, name, index=index )
        __expand_input_elems( repeat_elem, new_prefix )
        __pull_up_params( root_elem, repeat_elem )
        root_elem.remove( repeat_elem )

    cond_elems = root_elem.findall( 'conditional' )
    for cond_elem in cond_elems:
        new_prefix = __prefix_join( prefix, cond_elem.get( "name" ) )
        __expand_input_elems( cond_elem, new_prefix )
        __pull_up_params( root_elem, cond_elem )
        root_elem.remove( cond_elem )

    section_elems = root_elem.findall( 'section' )
    for section_elem in section_elems:
        new_prefix = __prefix_join( prefix, section_elem.get( "name" ) )
        __expand_input_elems( section_elem, new_prefix )
        __pull_up_params( root_elem, section_elem )
        root_elem.remove( section_elem )


def __append_prefix_to_params( elem, prefix ):
    for param_elem in elem.findall( 'param' ):
        param_elem.set( "name", __prefix_join( prefix, param_elem.get( "name" ) ) )


def __pull_up_params( parent_elem, child_elem ):
    for param_elem in child_elem.findall( 'param' ):
        parent_elem.append( param_elem )
        child_elem.remove( param_elem )


def __prefix_join( prefix, name, index=None ):
    name = name if index is None else "%s_%d" % ( name, index )
    return name if not prefix else "%s|%s" % ( prefix, name )


def _copy_to_dict_if_present( elem, rval, attributes ):
    for attribute in attributes:
        if attribute in elem.attrib:
            rval[attribute] = elem.get(attribute)
    return rval


def __parse_inputs_elems( test_elem, i ):
    raw_inputs = []
    for param_elem in test_elem.findall( "param" ):
        name, value, attrib = __parse_param_elem( param_elem, i )
        raw_inputs.append( ( name, value, attrib ) )
    return raw_inputs


def __parse_param_elem( param_elem, i=0 ):
    attrib = dict( param_elem.attrib )
    if 'values' in attrib:
        value = attrib[ 'values' ].split( ',' )
    elif 'value' in attrib:
        value = attrib['value']
    else:
        value = None
    attrib['children'] = list( param_elem.getchildren() )
    if attrib['children']:
        # At this time, we can assume having children only
        # occurs on DataToolParameter test items but this could
        # change and would cause the below parsing to change
        # based upon differences in children items
        attrib['metadata'] = []
        attrib['composite_data'] = []
        attrib['edit_attributes'] = []
        # Composite datasets need to be renamed uniquely
        composite_data_name = None
        for child in attrib['children']:
            if child.tag == 'composite_data':
                attrib['composite_data'].append( child )
                if composite_data_name is None:
                    # Generate a unique name; each test uses a
                    # fresh history.
                    composite_data_name = '_COMPOSITE_RENAMED_t%d_%s' \
                        % ( i, uuid.uuid1().hex )
            elif child.tag == 'metadata':
                attrib['metadata'].append( child )
            elif child.tag == 'metadata':
                attrib['metadata'].append( child )
            elif child.tag == 'edit_attributes':
                attrib['edit_attributes'].append( child )
            elif child.tag == 'collection':
                attrib[ 'collection' ] = TestCollectionDef( child, __parse_param_elem )
        if composite_data_name:
            # Composite datasets need implicit renaming;
            # inserted at front of list so explicit declarations
            # take precedence
            attrib['edit_attributes'].insert( 0, { 'type': 'name', 'value': composite_data_name } )
    name = attrib.pop( 'name' )
    return ( name, value, attrib )


class StdioParser(object):

    def __init__(self, root):
        try:
            self.stdio_exit_codes = list()
            self.stdio_regexes = list()

            # We should have a single <stdio> element, but handle the case for
            # multiples.
            # For every stdio element, add all of the exit_code and regex
            # subelements that we find:
            for stdio_elem in ( root.findall( 'stdio' ) ):
                self.parse_stdio_exit_codes( stdio_elem )
                self.parse_stdio_regexes( stdio_elem )
        except Exception:
            log.error( "Exception in parse_stdio! " + str(sys.exc_info()) )

    def parse_stdio_exit_codes( self, stdio_elem ):
        """
        Parse the tool's <stdio> element's <exit_code> subelements.
        This will add all of those elements, if any, to self.stdio_exit_codes.
        """
        try:
            # Look for all <exit_code> elements. Each exit_code element must
            # have a range/value.
            # Exit-code ranges have precedence over a single exit code.
            # So if there are value and range attributes, we use the range
            # attribute. If there is neither a range nor a value, then print
            # a warning and skip to the next.
            for exit_code_elem in ( stdio_elem.findall( "exit_code" ) ):
                exit_code = ToolStdioExitCode()
                # Each exit code has an optional description that can be
                # part of the "desc" or "description" attributes:
                exit_code.desc = exit_code_elem.get( "desc" )
                if exit_code.desc is None:
                    exit_code.desc = exit_code_elem.get( "description" )
                # Parse the error level:
                exit_code.error_level = (
                    self.parse_error_level( exit_code_elem.get( "level" )))
                code_range = exit_code_elem.get( "range", "" )
                if code_range is None:
                    code_range = exit_code_elem.get( "value", "" )
                if code_range is None:
                    log.warning( "Tool stdio exit codes must have " +
                                 "a range or value" )
                    continue
                # Parse the range. We look for:
                #   :Y
                #  X:
                #  X:Y   - Split on the colon. We do not allow a colon
                #          without a beginning or end, though we could.
                # Also note that whitespace is eliminated.
                # TODO: Turn this into a single match - it should be
                # more efficient.
                code_range = re.sub( "\s", "", code_range )
                code_ranges = re.split( ":", code_range )
                if ( len( code_ranges ) == 2 ):
                    if ( code_ranges[0] is None or '' == code_ranges[0] ):
                        exit_code.range_start = float( "-inf" )
                    else:
                        exit_code.range_start = int( code_ranges[0] )
                    if ( code_ranges[1] is None or '' == code_ranges[1] ):
                        exit_code.range_end = float( "inf" )
                    else:
                        exit_code.range_end = int( code_ranges[1] )
                # If we got more than one colon, then ignore the exit code.
                elif ( len( code_ranges ) > 2 ):
                    log.warning( "Invalid tool exit_code range %s - ignored"
                                 % code_range )
                    continue
                # Else we have a singular value. If it's not an integer, then
                # we'll just write a log message and skip this exit_code.
                else:
                    try:
                        exit_code.range_start = int( code_range )
                    except:
                        log.error( code_range )
                        log.warning( "Invalid range start for tool's exit_code %s: exit_code ignored" % code_range )
                        continue
                    exit_code.range_end = exit_code.range_start
                # TODO: Check if we got ">", ">=", "<", or "<=":
                # Check that the range, regardless of how we got it,
                # isn't bogus. If we have two infinite values, then
                # the start must be -inf and the end must be +inf.
                # So at least warn about this situation:
                if ( isinf( exit_code.range_start ) and
                     isinf( exit_code.range_end ) ):
                    log.warning( "Tool exit_code range %s will match on " +
                                 "all exit codes" % code_range )
                self.stdio_exit_codes.append( exit_code )
        except Exception:
            log.error( "Exception in parse_stdio_exit_codes! " +
                       str(sys.exc_info()) )
            trace = sys.exc_info()[2]
            if trace is not None:
                trace_msg = repr( traceback.format_tb( trace ) )
                log.error( "Traceback: %s" % trace_msg )

    def parse_stdio_regexes( self, stdio_elem ):
        """
        Look in the tool's <stdio> elem for all <regex> subelements
        that define how to look for warnings and fatal errors in
        stdout and stderr. This will add all such regex elements
        to the Tols's stdio_regexes list.
        """
        try:
            # Look for every <regex> subelement. The regular expression
            # will have "match" and "source" (or "src") attributes.
            for regex_elem in ( stdio_elem.findall( "regex" ) ):
                # TODO: Fill in ToolStdioRegex
                regex = ToolStdioRegex()
                # Each regex has an optional description that can be
                # part of the "desc" or "description" attributes:
                regex.desc = regex_elem.get( "desc" )
                if regex.desc is None:
                    regex.desc = regex_elem.get( "description" )
                # Parse the error level
                regex.error_level = (
                    self.parse_error_level( regex_elem.get( "level" ) ) )
                regex.match = regex_elem.get( "match", "" )
                if regex.match is None:
                    # TODO: Convert the offending XML element to a string
                    log.warning( "Ignoring tool's stdio regex element %s - "
                                 "the 'match' attribute must exist" )
                    continue
                # Parse the output sources. We look for the "src", "source",
                # and "sources" attributes, in that order. If there is no
                # such source, then the source defaults to stderr & stdout.
                # Look for a comma and then look for "err", "error", "out",
                # and "output":
                output_srcs = regex_elem.get( "src" )
                if output_srcs is None:
                    output_srcs = regex_elem.get( "source" )
                if output_srcs is None:
                    output_srcs = regex_elem.get( "sources" )
                if output_srcs is None:
                    output_srcs = "output,error"
                output_srcs = re.sub( "\s", "", output_srcs )
                src_list = re.split( ",", output_srcs )
                # Just put together anything to do with "out", including
                # "stdout", "output", etc. Repeat for "stderr", "error",
                # and anything to do with "err". If neither stdout nor
                # stderr were specified, then raise a warning and scan both.
                for src in src_list:
                    if re.search( "both", src, re.IGNORECASE ):
                        regex.stdout_match = True
                        regex.stderr_match = True
                    if re.search( "out", src, re.IGNORECASE ):
                        regex.stdout_match = True
                    if re.search( "err", src, re.IGNORECASE ):
                        regex.stderr_match = True
                    if (not regex.stdout_match and not regex.stderr_match):
                        log.warning( "Tool id %s: unable to determine if tool "
                                     "stream source scanning is output, error, "
                                     "or both. Defaulting to use both." % self.id )
                        regex.stdout_match = True
                        regex.stderr_match = True
                self.stdio_regexes.append( regex )
        except Exception:
            log.error( "Exception in parse_stdio_exit_codes! " +
                       str(sys.exc_info()) )
            trace = sys.exc_info()[2]
            if trace is not None:
                trace_msg = repr( traceback.format_tb( trace ) )
                log.error( "Traceback: %s" % trace_msg )

    # TODO: This method doesn't have to be part of the Tool class.
    def parse_error_level( self, err_level ):
        """
        Parses error level and returns error level enumeration. If
        unparsable, returns 'fatal'
        """
        from galaxy.jobs.error_level import StdioErrorLevel
        return_level = StdioErrorLevel.FATAL
        try:
            if err_level:
                if ( re.search( "log", err_level, re.IGNORECASE ) ):
                    return_level = StdioErrorLevel.LOG
                elif ( re.search( "warning", err_level, re.IGNORECASE ) ):
                    return_level = StdioErrorLevel.WARNING
                elif ( re.search( "fatal", err_level, re.IGNORECASE ) ):
                    return_level = StdioErrorLevel.FATAL
                else:
                    log.debug( "Tool %s: error level %s did not match log/warning/fatal" %
                               ( self.id, err_level ) )
        except Exception:
            log.error( "Exception in parse_error_level " +
                       str(sys.exc_info() ) )
            trace = sys.exc_info()[2]
            if trace is not None:
                trace_msg = repr( traceback.format_tb( trace ) )
                log.error( "Traceback: %s" % trace_msg )
        return return_level


class XmlPagesSource(PagesSource):

    def __init__(self, root):
        self.input_elem = root.find("inputs")
        page_sources = []
        if self.input_elem is not None:
            pages_elem = self.input_elem.findall( "page" )
            for page in ( pages_elem or [ self.input_elem ] ):
                page_sources.append(XmlPageSource(page))
        super(XmlPagesSource, self).__init__(page_sources)

    @property
    def inputs_defined(self):
        return self.input_elem is not None


class XmlPageSource(PageSource):

    def __init__(self, parent_elem):
        self.parent_elem = parent_elem

    def parse_display(self):
        display_elem = self.parent_elem.find("display")
        if display_elem is not None:
            display = xml_to_string(display_elem)
        else:
            display = None
        return display

    def parse_input_sources(self):
        return map(XmlInputSource, self.parent_elem)


class XmlInputSource(InputSource):

    def __init__(self, input_elem):
        self.input_elem = input_elem
        self.input_type = self.input_elem.tag

    def parse_input_type(self):
        return self.input_type

    def elem(self):
        return self.input_elem

    def get(self, key, value=None):
        return self.input_elem.get(key, value)

    def get_bool(self, key, default):
        return string_as_bool( self.get(key, default ) )

    def parse_label(self):
        return xml_text(self.input_elem, "label")

    def parse_help(self):
        return xml_text(self.input_elem, "help")

    def parse_sanitizer_elem(self):
        return self.input_elem.find("sanitizer")

    def parse_validator_elems(self):
        return self.input_elem.findall("validator")

    def parse_dynamic_options_elem(self):
        """ Return a galaxy.tools.parameters.dynamic_options.DynamicOptions
        if appropriate.
        """
        options_elem = self.input_elem.find( 'options' )
        return options_elem

    def parse_static_options(self):
        static_options = list()
        elem = self.input_elem
        for index, option in enumerate( elem.findall( "option" ) ):
            value = option.get( "value" )
            selected = string_as_bool( option.get( "selected", False ) )
            static_options.append( ( option.text or value, value, selected ) )
        return static_options

    def parse_optional(self, default=None):
        """ Return boolean indicating wheter parameter is optional. """
        elem = self.input_elem
        if self.get('type') == "data_column":
            # Allow specifing force_select for backward compat., but probably
            # should use optional going forward for consistency with other
            # parameters.
            if "force_select" in elem.attrib:
                force_select = string_as_bool( elem.get( "force_select" ) )
            else:
                force_select = not string_as_bool( elem.get( "optional", False ) )
            return not force_select

        if default is None:
            default = self.default_optional
        return self.get_bool( "optional", default )

    def parse_conversion_tuples(self):
        elem = self.input_elem
        conversions = []
        for conv_elem in elem.findall( "conversion" ):
            name = conv_elem.get( "name" )  # name for commandline substitution
            conv_extensions = conv_elem.get( "type" )  # target datatype extension
            conversions.append((name, conv_extensions))
        return conversions

    def parse_nested_inputs_source(self):
        elem = self.input_elem
        return XmlPageSource(elem)

    def parse_test_input_source(self):
        elem = self.input_elem
        input_elem = elem.find( "param" )
        assert input_elem is not None, "<conditional> must have a child <param>"
        return XmlInputSource(input_elem)

    def parse_when_input_sources(self):
        elem = self.input_elem

        sources = []
        for case_elem in elem.findall( "when" ):
            value = case_elem.get( "value" )
            case_page_source = XmlPageSource(case_elem)
            sources.append((value, case_page_source))
        return sources



"""
Support for dynamically modifying output attributes.
"""

import logging
import os.path
import re
import string

from galaxy import util

log = logging.getLogger( __name__ )

# Attributes tool developer may want to query on dataset collections.
COLLECTION_ATTRIBUTES = [ "collection_type" ]


class ToolOutputActionGroup( object ):
    """
    Manages a set of tool output dataset actions directives
    """
    tag = "group"

    def __init__( self, parent, config_elem ):
        self.parent = parent
        self.actions = []
        if config_elem is not None:
            for elem in config_elem:
                if elem.tag == "conditional":
                    self.actions.append( ToolOutputActionConditional( self, elem ) )
                elif elem.tag == "action":
                    self.actions.append( ToolOutputAction.from_elem( self, elem ) )
                else:
                    log.debug( "Unknown ToolOutputAction tag specified: %s" % elem.tag )

    def apply_action( self, output_dataset, other_values ):
        for action in self.actions:
            action.apply_action( output_dataset, other_values )

    @property
    def tool( self ):
        return self.parent.tool

    def __len__( self ):
        return len( self.actions )


class ToolOutputActionConditionalWhen( ToolOutputActionGroup ):
    tag = "when"

    @classmethod
    def from_elem( cls, parent, when_elem ):
        """Loads the proper when by attributes of elem"""
        when_value = when_elem.get( "value", None )
        if when_value is not None:
            return ValueToolOutputActionConditionalWhen( parent, when_elem, when_value )
        else:
            when_value = when_elem.get( "datatype_isinstance", None )
            if when_value is not None:
                return DatatypeIsInstanceToolOutputActionConditionalWhen( parent, when_elem, when_value )
        raise TypeError( "When type not implemented" )

    def __init__( self, parent, config_elem, value ):
        super( ToolOutputActionConditionalWhen, self ).__init__( parent, config_elem )
        self.value = value

    def is_case( self, output_dataset, other_values ):
        raise TypeError( "Not implemented" )

    def get_ref( self, output_dataset, other_values ):
        ref = other_values
        for ref_name in self.parent.name:
            assert ref_name in ref, "Required dependency '%s' not found in incoming values" % ref_name
            ref = ref.get( ref_name )
        return ref

    def apply_action( self, output_dataset, other_values ):
        if self.is_case( output_dataset, other_values ):
            return super( ToolOutputActionConditionalWhen, self ).apply_action( output_dataset, other_values )


class ValueToolOutputActionConditionalWhen( ToolOutputActionConditionalWhen ):
    tag = "when value"

    def is_case( self, output_dataset, other_values ):
        ref = self.get_ref( output_dataset, other_values )
        return bool( str( ref ) == self.value )


class DatatypeIsInstanceToolOutputActionConditionalWhen( ToolOutputActionConditionalWhen ):
    tag = "when datatype_isinstance"

    def __init__( self, parent, config_elem, value ):
        super( DatatypeIsInstanceToolOutputActionConditionalWhen, self ).__init__( parent, config_elem, value )
        self.value = type( self.tool.app.datatypes_registry.get_datatype_by_extension( value ) )

    def is_case( self, output_dataset, other_values ):
        ref = self.get_ref( output_dataset, other_values )
        return isinstance( ref.datatype, self.value )


class ToolOutputActionConditional( object ):
    tag = "conditional"

    def __init__( self, parent, config_elem ):
        self.parent = parent
        self.name = config_elem.get( 'name', None )
        assert self.name is not None, "Required 'name' attribute missing from ToolOutputActionConditional"
        self.name = self.name.split( '.' )
        self.cases = []
        for when_elem in config_elem.findall( 'when' ):
            self.cases.append( ToolOutputActionConditionalWhen.from_elem( self, when_elem ) )

    def apply_action( self, output_dataset, other_values ):
        for case in self.cases:
            case.apply_action( output_dataset, other_values )

    @property
    def tool( self ):
        return self.parent.tool


class ToolOutputAction( object ):
    tag = "action"

    @classmethod
    def from_elem( cls, parent, elem ):
        """Loads the proper action by the type attribute of elem"""
        action_type = elem.get( 'type', None )
        assert action_type is not None, "Required 'type' attribute missing from ToolOutputAction"
        return action_types[ action_type ]( parent, elem )

    def __init__( self, parent, elem ):
        self.parent = parent
        self.default = elem.get( 'default', None )
        option_elem = elem.find( 'option' )
        self.option = ToolOutputActionOption.from_elem( self, option_elem )

    def apply_action( self, output_dataset, other_values ):
        raise TypeError( "Not implemented" )

    @property
    def tool( self ):
        return self.parent.tool


class ToolOutputActionOption( object ):
    tag = "object"

    @classmethod
    def from_elem( cls, parent, elem ):
        """Loads the proper action by the type attribute of elem"""
        if elem is None:
            option_type = NullToolOutputActionOption.tag  # no ToolOutputActionOption's have been defined, use implicit NullToolOutputActionOption
        else:
            option_type = elem.get( 'type', None )
        assert option_type is not None, "Required 'type' attribute missing from ToolOutputActionOption"
        return option_types[ option_type ]( parent, elem )

    def __init__( self, parent, elem ):
        self.parent = parent
        self.filters = []
        if elem is not None:
            for filter_elem in elem.findall( 'filter' ):
                self.filters.append( ToolOutputActionOptionFilter.from_elem( self, filter_elem ) )

    def get_value( self, other_values ):
        raise TypeError( "Not implemented" )

    @property
    def tool( self ):
        return self.parent.tool


class NullToolOutputActionOption( ToolOutputActionOption ):
    tag = "null_option"

    def get_value( self, other_values ):
        return None


class FromFileToolOutputActionOption( ToolOutputActionOption ):
    tag = "from_file"

    def __init__( self, parent, elem ):
        super( FromFileToolOutputActionOption, self ).__init__( parent, elem )
        self.name = elem.get( 'name', None )
        assert self.name is not None, "Required 'name' attribute missing from FromFileToolOutputActionOption"
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from FromFileToolOutputActionOption"
        self.column = int( self.column )
        self.offset = elem.get( 'offset', -1 )
        self.offset = int( self.offset )
        self.separator = elem.get( 'separator', '\t' )
        self.options = []
        data_file = self.name
        if not os.path.isabs( data_file ):
            data_file = os.path.join( self.tool.app.config.tool_data_path, data_file )
        for line in open( data_file ):
            self.options.append( line.rstrip( '\n\r' ).split( self.separator ) )

    def get_value( self, other_values ):
        options = self.options
        for filter in self.filters:
            options = filter.filter_options( options, other_values )
        try:
            if options:
                return str( options[ self.offset ][ self.column ] )
        except Exception as e:
            log.debug( "Error in FromFileToolOutputActionOption get_value: %s" % e )
        return None


class FromParamToolOutputActionOption( ToolOutputActionOption ):
    tag = "from_param"

    def __init__( self, parent, elem ):
        super( FromParamToolOutputActionOption, self ).__init__( parent, elem )
        self.name = elem.get( 'name', None )
        assert self.name is not None, "Required 'name' attribute missing from FromFileToolOutputActionOption"
        self.name = self.name.split( '.' )
        self.column = elem.get( 'column', 0 )
        self.column = int( self.column )
        self.offset = elem.get( 'offset', -1 )
        self.offset = int( self.offset )
        self.param_attribute = elem.get( 'param_attribute', [] )
        if self.param_attribute:
            self.param_attribute = self.param_attribute.split( '.' )

    def get_value( self, other_values ):
        value = other_values
        for ref_name in self.name:
            assert ref_name in value, "Required dependency '%s' not found in incoming values" % ref_name
            value = value.get( ref_name )
        for attr_name in self.param_attribute:
            # if the value is a list from a repeat tag you can access the first element of the repeat with
            # artifical 'first' attribute_name. For example: .. param_attribute="first.input_mate1.ext"
            if isinstance(value, list) and attr_name == 'first':
                value = value[0]
            elif isinstance(value, dict):
                value = value[ attr_name ]
            elif hasattr( value, "collection" ) and value not in COLLECTION_ATTRIBUTES:
                # if this is an HDCA for instance let reverse.ext grab
                # the reverse element and then continue for loop to grab
                # dataset extension
                value = value.collection[ attr_name ].element_object
            elif hasattr( value, "collection" ) and value in COLLECTION_ATTRIBUTES:
                value = getattr( value.collection, attr_name )
            else:
                value = getattr( value, attr_name )
        options = [ [ str( value ) ] ]
        for filter in self.filters:
            options = filter.filter_options( options, other_values )
        try:
            if options:
                return str( options[ self.offset ][ self.column ] )
        except Exception as e:
            log.debug( "Error in FromParamToolOutputActionOption get_value: %s" % e )
        return None


class FromDataTableOutputActionOption( ToolOutputActionOption ):
    tag = "from_data_table"

    # TODO: allow accessing by column 'name' not just index
    def __init__( self, parent, elem ):
        super( FromDataTableOutputActionOption, self ).__init__( parent, elem )
        self.name = elem.get( 'name', None )
        assert self.name is not None, "Required 'name' attribute missing from FromDataTableOutputActionOption"
        self.missing_tool_data_table_name = None
        if self.name in self.tool.app.tool_data_tables:
            self.options = self.tool.app.tool_data_tables[ self.name ].get_fields()
            self.column = elem.get( 'column', None )
            assert self.column is not None, "Required 'column' attribute missing from FromDataTableOutputActionOption"
            self.column = int( self.column )
            self.offset = elem.get( 'offset', -1 )
            self.offset = int( self.offset )
        else:
            self.options = []
            self.missing_tool_data_table_name = self.name

    def get_value( self, other_values ):
        if self.options:
            options = self.options
        else:
            options = []
        for filter in self.filters:
            options = filter.filter_options( options, other_values )
        try:
            if options:
                return str( options[ self.offset ][ self.column ] )
        except Exception as e:
            log.debug( "Error in FromDataTableOutputActionOption get_value: %s" % e )
        return None


class MetadataToolOutputAction( ToolOutputAction ):
    tag = "metadata"

    def __init__( self, parent, elem ):
        super( MetadataToolOutputAction, self ).__init__( parent, elem )
        self.name = elem.get( 'name', None )
        assert self.name is not None, "Required 'name' attribute missing from MetadataToolOutputAction"

    def apply_action( self, output_dataset, other_values ):
        value = self.option.get_value( other_values )
        if value is None and self.default is not None:
            value = self.default
        if value is not None:
            setattr( output_dataset.metadata, self.name, value )


class FormatToolOutputAction( ToolOutputAction ):
    tag = "format"

    def __init__( self, parent, elem ):
        super( FormatToolOutputAction, self ).__init__( parent, elem )
        self.default = elem.get( 'default', None )

    def apply_action( self, output_dataset, other_values ):
        value = self.option.get_value( other_values )
        if value is None and self.default is not None:
            value = self.default
        if value is not None:
            output_dataset.extension = value


class ToolOutputActionOptionFilter( object ):
    tag = "filter"

    @classmethod
    def from_elem( cls, parent, elem ):
        """Loads the proper action by the type attribute of elem"""
        filter_type = elem.get( 'type', None )
        assert filter_type is not None, "Required 'type' attribute missing from ToolOutputActionOptionFilter"
        return filter_types[ filter_type ]( parent, elem )

    def __init__( self, parent, elem ):
        self.parent = parent

    def filter_options( self, options, other_values ):
        raise TypeError( "Not implemented" )

    @property
    def tool( self ):
        return self.parent.tool


class ParamValueToolOutputActionOptionFilter( ToolOutputActionOptionFilter ):
    tag = "param_value"

    def __init__( self, parent, elem ):
        super( ParamValueToolOutputActionOptionFilter, self ).__init__( parent, elem )
        self.ref = elem.get( 'ref', None )
        if self.ref:
            self.ref = self.ref.split( '.' )
        self.value = elem.get( 'value', None )
        assert self.ref != self.value, "Required 'ref' or 'value' attribute missing from ParamValueToolOutputActionOptionFilter"
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from ParamValueToolOutputActionOptionFilter"
        self.column = int( self.column )
        self.keep = util.string_as_bool( elem.get( "keep", 'True' ) )
        self.compare = parse_compare_type( elem.get( 'compare', None ) )
        self.cast = parse_cast_attribute( elem.get( "cast", None ) )
        self.param_attribute = elem.get( 'param_attribute', [] )
        if self.param_attribute:
            self.param_attribute = self.param_attribute.split( '.' )

    def filter_options( self, options, other_values ):
        if self.ref:
            # find ref value
            value = other_values
            for ref_name in self.ref:
                assert ref_name in value, "Required dependency '%s' not found in incoming values" % ref_name
                value = value.get( ref_name )
            for attr_name in self.param_attribute:
                value = getattr( value, attr_name )
            value = str( value )
        else:
            value = self.value
        value = self.cast( value )
        rval = []
        for fields in options:
            try:
                if self.keep == ( self.compare( self.cast( fields[self.column] ), value ) ):
                    rval.append( fields )
            except Exception as e:
                log.debug(e)
                continue  # likely a bad cast or column out of range
        return rval


class InsertColumnToolOutputActionOptionFilter( ToolOutputActionOptionFilter ):
    tag = "insert_column"

    def __init__( self, parent, elem ):
        super( InsertColumnToolOutputActionOptionFilter, self ).__init__( parent, elem )
        self.ref = elem.get( 'ref', None )
        if self.ref:
            self.ref = self.ref.split( '.' )
        self.value = elem.get( 'value', None )
        assert self.ref != self.value, "Required 'ref' or 'value' attribute missing from InsertColumnToolOutputActionOptionFilter"
        self.column = elem.get( 'column', None )  # None is append
        if self.column:
            self.column = int( self.column )
        self.iterate = util.string_as_bool( elem.get( "iterate", 'False' ) )

    def filter_options( self, options, other_values ):
        if self.ref:
            # find ref value
            value = other_values
            for ref_name in self.ref:
                assert ref_name in value, "Required dependency '%s' not found in incoming values" % ref_name
                value = value.get( ref_name )
            value = str( value )
        else:
            value = self.value
        if self.iterate:
            value = int( value )
        rval = []
        for fields in options:
            if self.column is None:
                rval.append( fields + [ str( value ) ] )
            else:
                fields = list( fields )
                fields.insert( self.column, str( value ) )
                rval.append( fields )
            if self.iterate:
                value += 1
        return rval


class MultipleSplitterFilter( ToolOutputActionOptionFilter ):
    tag = "multiple_splitter"

    def __init__( self, parent, elem ):
        super( MultipleSplitterFilter, self ).__init__( parent, elem )
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from MultipleSplitterFilter"
        self.column = int( self.column )
        self.separator = elem.get( "separator", "," )

    def filter_options( self, options, other_values ):
        rval = []
        for fields in options:
            for field in fields[self.column].split( self.separator ):
                rval.append( fields[0:self.column] + [field] + fields[self.column + 1:] )
        return rval


class ColumnStripFilter( ToolOutputActionOptionFilter ):
    tag = "column_strip"

    def __init__( self, parent, elem ):
        super( ColumnStripFilter, self ).__init__( parent, elem )
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from ColumnStripFilter"
        self.column = int( self.column )
        self.strip = elem.get( "strip", None )

    def filter_options( self, options, other_values ):
        rval = []
        for fields in options:
            rval.append( fields[0:self.column] + [ fields[self.column].strip( self.strip ) ] + fields[self.column + 1:] )
        return rval


class ColumnReplaceFilter( ToolOutputActionOptionFilter ):
    tag = "column_replace"

    def __init__( self, parent, elem ):
        super( ColumnReplaceFilter, self ).__init__( parent, elem )
        self.old_column = elem.get( 'old_column', None )
        self.old_value = elem.get( "old_value", None )
        self.new_value = elem.get( "new_value", None )
        self.new_column = elem.get( 'new_column', None )
        assert ( bool( self.old_column ) ^ bool( self.old_value ) and bool( self.new_column ) ^ bool( self.new_value ) ), "Required 'old_column' or 'old_value' and 'new_column' or 'new_value' attribute missing from ColumnReplaceFilter"
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from ColumnReplaceFilter"
        self.column = int( self.column )
        if self.old_column is not None:
            self.old_column = int( self.old_column )
        if self.new_column is not None:
            self.new_column = int( self.new_column )

    def filter_options( self, options, other_values ):
        rval = []
        for fields in options:
            if self.old_column:
                old_value = fields[self.old_column]
            else:
                old_value = self.old_value
            if self.new_column:
                new_value = fields[self.new_column]
            else:
                new_value = self.new_value
            rval.append( fields[0:self.column] + [ fields[self.column].replace( old_value, new_value ) ] + fields[self.column + 1:] )
        return rval


class MetadataValueFilter( ToolOutputActionOptionFilter ):
    tag = "metadata_value"

    def __init__( self, parent, elem ):
        super( MetadataValueFilter, self ).__init__( parent, elem )
        self.ref = elem.get( 'ref', None )
        assert self.ref is not None, "Required 'ref' attribute missing from MetadataValueFilter"
        self.ref = self.ref.split( '.' )
        self.name = elem.get( 'name', None )
        assert self.name is not None, "Required 'name' attribute missing from MetadataValueFilter"
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from MetadataValueFilter"
        self.column = int( self.column )
        self.keep = util.string_as_bool( elem.get( "keep", 'True' ) )
        self.compare = parse_compare_type( elem.get( 'compare', None ) )

    def filter_options( self, options, other_values ):
        ref = other_values
        for ref_name in self.ref:
            assert ref_name in ref, "Required dependency '%s' not found in incoming values" % ref_name
            ref = ref.get( ref_name )
        value = str( getattr( ref.metadata, self.name ) )
        rval = []
        for fields in options:
            if self.keep == ( self.compare( fields[self.column], value ) ):
                rval.append( fields )
        return rval


class BooleanFilter( ToolOutputActionOptionFilter ):
    tag = "boolean"

    def __init__( self, parent, elem ):
        super( BooleanFilter, self ).__init__( parent, elem )
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from BooleanFilter"
        self.column = int( self.column )
        self.keep = util.string_as_bool( elem.get( "keep", 'True' ) )
        self.cast = parse_cast_attribute( elem.get( "cast", None ) )

    def filter_options( self, options, other_values ):
        rval = []
        for fields in options:
            try:
                value = fields[self.column]
                value = self.cast( value )
            except:
                value = False  # unable to cast or access value; treat as false
            if self.keep == bool( value ):
                rval.append( fields )
        return rval


class StringFunctionFilter( ToolOutputActionOptionFilter ):
    tag = "string_function"

    def __init__( self, parent, elem ):
        super( StringFunctionFilter, self ).__init__( parent, elem )
        self.column = elem.get( 'column', None )
        assert self.column is not None, "Required 'column' attribute missing from StringFunctionFilter"
        self.column = int( self.column )
        self.function = elem.get( "name", None )
        assert self.function in [ 'lower', 'upper' ], "Required function 'name' missing or invalid from StringFunctionFilter"  # add function names as needed
        self.function = getattr( string, self.function )

    def filter_options( self, options, other_values ):
        rval = []
        for fields in options:
            rval.append( fields[0:self.column] + [ self.function( fields[self.column] ) ] + fields[self.column + 1:] )
        return rval


# tag to class lookups
action_types = {}
for action_type in [ MetadataToolOutputAction, FormatToolOutputAction ]:
    action_types[ action_type.tag ] = action_type

option_types = {}
for option_type in [ NullToolOutputActionOption, FromFileToolOutputActionOption, FromParamToolOutputActionOption, FromDataTableOutputActionOption ]:
    option_types[ option_type.tag ] = option_type

filter_types = {}
for filter_type in [ ParamValueToolOutputActionOptionFilter, InsertColumnToolOutputActionOptionFilter, MultipleSplitterFilter, ColumnStripFilter, MetadataValueFilter, BooleanFilter, StringFunctionFilter, ColumnReplaceFilter ]:
    filter_types[ filter_type.tag ] = filter_type


# helper classes
# determine cast function
def parse_cast_attribute( cast ):
    if cast == 'string_as_bool':
        cast = util.string_as_bool
    elif cast == 'int':
        cast = int
    elif cast == 'str':
        cast = str
    else:
        # return value as-is
        def cast(x):
            return x
    return cast


# comparison
def parse_compare_type( compare ):
    if compare is None:
        compare = 'eq'
    assert compare in compare_types, "Invalid compare type specified: %s" % compare
    return compare_types[ compare ]


def compare_eq( value1, value2  ):
    return value1 == value2


def compare_neq( value1, value2  ):
    return value1 != value2


def compare_gt( value1, value2  ):
    return value1 > value2


def compare_gte( value1, value2  ):
    return value1 >= value2


def compare_lt( value1, value2  ):
    return value1 < value2


def compare_lte( value1, value2  ):
    return value1 <= value2


def compare_in( value1, value2 ):
    return value1 in value2


def compare_startswith( value1, value2 ):
    return value1.startswith( value2 )


def compare_endswith( value1, value2 ):
    return value1.endswith( value2 )


def compare_re_search( value1, value2 ):
    # checks pattern=value2 in value1
    return bool( re.search( value2, value1 ) )


compare_types = {
    'eq': compare_eq,
    'neq': compare_neq,
    'gt': compare_gt,
    'gte': compare_gte,
    'lt': compare_lt,
    'lte': compare_lte,
    'in': compare_in,
    'startswith': compare_startswith,
    'endswith': compare_endswith,
    "re_search": compare_re_search
}



from galaxy.exceptions import (
    NotImplemented,
    RequestParameterMissingException
)


class DependencyResolversView(object):
    """ Provide a RESTfulish/JSONy interface to a galaxy.tools.deps.DependencyResolver
    object. This can be adapted by the Galaxy web framework or other web apps.
    """

    def __init__(self, app):
        self._app = app

    def index(self):
        return [r.to_dict() for r in self._dependency_resolvers]

    def show(self, index):
        return self._dependency_resolver(index).to_dict()

    def reload(self):
        self.toolbox.reload_dependency_manager()

    def manager_requirements(self):
        requirements = []
        for index, resolver in enumerate(self._dependency_resolvers):
            if not hasattr(resolver, "list_dependencies"):
                continue
            for requirement in resolver.list_dependencies():
                requirements.append({"index": index, "requirement": requirement.to_dict()})
        return requirements

    def resolver_requirements(self, index):
        requirements = []
        resolver = self._dependency_resolver(index)
        if not hasattr(resolver, "list_dependencies"):
            raise NotImplemented()
        for requirement in resolver.list_dependencies():
            requirements.append(requirement.to_dict())
        return requirements

    def manager_dependency(self, **kwds):
        return self._dependency(**kwds)

    def resolver_dependency(self, index, **kwds):
        return self._dependency(**kwds)

    def show_dependencies(self, tool_requirements_d, installed_tool_dependencies=None):
        """
        Resolves dependencies to build a requirements status in the admin panel/API
        """
        kwds = {'install': False,
                'return_null': True,
                'installed_tool_dependencies': installed_tool_dependencies}
        dependencies_per_tool = {tool: self._dependency_manager.requirements_to_dependencies(requirements, **kwds) for tool, requirements in tool_requirements_d.items()}
        return dependencies_per_tool

    def install_dependencies(self, requirements):
        return self._dependency_manager._requirements_to_dependencies_dict(requirements, **{'install': True})

    def install_dependency(self, index=None, **payload):
        """
        Installs dependency using highest priority resolver that supports dependency installation
        (Currently only the conda resolver supports this). If index is given, attempt
        installation directly using the corresponding resolver.
        Returns True on success, False on failure.
        payload is dictionary that must container name, version and type,
        e.g. {'name': 'numpy', version='1.9.1', type='package'}
        """
        if index:
            return self._install_dependency(index, **payload)
        else:
            for index in self.installable_resolvers:
                success = self._install_dependency(index, **payload)
                if success:
                    return success
            return False

    def _install_dependency(self, index, **payload):
        """
        Resolver install dependency should return True when installation succeeds,
        False if not successful
        """
        resolver = self._dependency_resolver(index)
        if not hasattr(resolver, "install_dependency"):
            raise NotImplemented()

        name, version, type, extra_kwds = self._parse_dependency_info(payload)
        return resolver.install_dependency(
            name=name,
            version=version,
            type=type,
            **extra_kwds
        )

    def _dependency(self, index=None, **kwds):
        if index is not None:
            index = int(index)

        name, version, type, extra_kwds = self._parse_dependency_info(kwds)
        resolve_kwds = dict(
            job_directory=None,
            index=index,
            **extra_kwds
        )
        dependency = self._dependency_manager.find_dep(
            name, version=version, type=type, **resolve_kwds
        )
        return dependency.to_dict()

    def _parse_dependency_info(self, kwds):
        extra_kwds = kwds.copy()
        name = extra_kwds.pop("name", None)
        if name is None:
            raise RequestParameterMissingException("Missing 'name' parameter required for resolution.")
        version = extra_kwds.pop("version", None)
        type = extra_kwds.pop("type", "package")
        return name, version, type, extra_kwds

    def _dependency_resolver(self, index):
        index = int(index)
        return self._dependency_resolvers[index]

    @property
    def _dependency_manager(self):
        return self._app.toolbox.dependency_manager

    @property
    def _dependency_resolvers(self):
        dependency_manager = self._dependency_manager
        dependency_resolvers = dependency_manager.dependency_resolvers
        return dependency_resolvers

    @property
    def installable_resolvers(self):
        """
        List index for all active resolvers that have the 'install_dependency' attribute
        """
        return [index for index, resolver in enumerate(self._dependency_resolvers) if hasattr(resolver, "install_dependency") and not resolver.disabled ]

    def get_requirements_status(self, tool_requirements_d, installed_tool_dependencies=None):
        dependencies = self.show_dependencies(tool_requirements_d, installed_tool_dependencies)
        # dependencies is a dict keyed on tool_ids, value is a ToolRequirements object for that tool.
        # We use the union of resolvable ToolRequirements to get resolved dependencies without duplicates.
        requirements = [r.resolvable for r in tool_requirements_d.values()]
        flat_tool_requirements = set().union(*requirements)
        flat_dependencies = []
        for requirements_odict in dependencies.values():
            for requirement in requirements_odict:
                if requirement in flat_tool_requirements:
                    flat_dependencies.append(requirements_odict[requirement])
                    flat_tool_requirements.remove(requirement)
        return [d.to_dict() for d in flat_dependencies]

    def clean(self, index=None, **kwds):
        if index:
            resolver = self._dependency_resolver(index)
            if not hasattr(resolver, "clean"):
                raise NotImplemented()
            else:
                resolver.clean()
                return "OK"
        else:
            [resolver.clean(**kwds) for resolver in self._dependency_resolvers if hasattr(resolver, 'clean')]
            return "OK"



from galaxy.util.xml_macros import (
    imported_macro_paths,
    load,
    raw_tool_xml_tree,
    template_macro_params,
)

load_tool = load

__all__ = ("load_tool", "raw_tool_xml_tree", "imported_macro_paths", "template_macro_params")



"""
Ordered dictionary implementation.
"""

from six.moves import UserDict
dict_alias = dict


class odict(UserDict):
    """
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/107747

    This dictionary class extends UserDict to record the order in which items are
    added. Calling keys(), values(), items(), etc. will return results in this
    order.
    """
    def __init__( self, dict=None ):
        item = dict
        self._keys = []
        if isinstance(item, dict_alias):
            UserDict.__init__( self, item )
        else:
            UserDict.__init__( self, None )
        if isinstance(item, list):
            for (key, value) in item:
                self[key] = value

    def __delitem__( self, key ):
        UserDict.__delitem__( self, key )
        self._keys.remove( key )

    def __setitem__( self, key, item ):
        UserDict.__setitem__( self, key, item )
        if key not in self._keys:
            self._keys.append( key )

    def clear( self ):
        UserDict.clear( self )
        self._keys = []

    def copy(self):
        new = odict()
        new.update( self )
        return new

    def items( self ):
        return zip( self._keys, self.values() )

    def keys( self ):
        return self._keys[:]

    def popitem( self ):
        try:
            key = self._keys[-1]
        except IndexError:
            raise KeyError( 'dictionary is empty' )
        val = self[ key ]
        del self[ key ]
        return ( key, val )

    def setdefault( self, key, failobj=None ):
        if key not in self._keys:
            self._keys.append( key )
        return UserDict.setdefault( self, key, failobj )

    def update( self, dict ):
        for ( key, val ) in dict.items():
            self.__setitem__( key, val )

    def values( self ):
        return map( self.get, self._keys )

    def iterkeys( self ):
        return iter( self._keys )

    def itervalues( self ):
        for key in self._keys:
            yield self.get( key )

    def iteritems( self ):
        for key in self._keys:
            yield key, self.get( key )

    def __iter__( self ):
        for key in self._keys:
            yield key

    def reverse( self ):
        self._keys.reverse()

    def insert( self, index, key, item ):
        if key not in self._keys:
            self._keys.insert( index, key )
            UserDict.__setitem__( self, key, item )



import pickle

from six.moves import cStringIO as StringIO


class AliasUnpickler( pickle.Unpickler ):
    def __init__( self, aliases, *args, **kw):
        pickle.Unpickler.__init__( self, *args, **kw )
        self.aliases = aliases

    def find_class( self, module, name ):
        module, name = self.aliases.get((module, name), (module, name))
        return pickle.Unpickler.find_class( self, module, name )


class AliasPickleModule( object ):
    def __init__( self, aliases ):
        self.aliases = aliases

    def dump( self, obj, fileobj, protocol=0):
        return pickle.dump( obj, fileobj, protocol )

    def dumps( self, obj, protocol=0 ):
        return pickle.dumps( obj, protocol )

    def load( self, fileobj ):
        return AliasUnpickler( self.aliases, fileobj ).load()

    def loads( self, string ):
        fileobj = StringIO( string )
        return AliasUnpickler( self.aliases, fileobj ).load()



from galaxy.util import plugin_config


class ToolLocationFetcher(object):

    def __init__(self):
        self.resolver_classes = self.__resolvers_dict()

    def __resolvers_dict( self ):
        import galaxy.tools.locations
        return plugin_config.plugins_dict(galaxy.tools.locations, 'scheme')

    def to_tool_path(self, path_or_uri_like, **kwds):
        if "://" not in path_or_uri_like:
            path = path_or_uri_like
        else:
            uri_like = path_or_uri_like
            if ":" not in path_or_uri_like:
                raise Exception("Invalid URI passed to get_tool_source")
            scheme, rest = uri_like.split(":", 2)
            if scheme not in self.resolver_classes:
                raise Exception("Unknown tool scheme [%s] for URI [%s]" % (scheme, uri_like))
            path = self.resolver_classes[scheme]().get_tool_source_path(uri_like)

        return path



""" This module is responsible for converting between Galaxy's tool
input description and the CWL description for a job json. """

import json
import logging
import os

from six import string_types

from galaxy.exceptions import RequestParameterInvalidException
from galaxy.util import safe_makedirs, string_as_bool

log = logging.getLogger(__name__)

NOT_PRESENT = object()

GALAXY_TO_CWL_TYPES = {
    'integer': 'integer',
    'float': 'float',
    'data': 'File',
    'boolean': 'boolean',
    'text': 'text'
}


def to_cwl_job(tool, param_dict, local_working_directory):
    """ tool is Galaxy's representation of the tool and param_dict is the
    parameter dictionary with wrapped values.
    """
    inputs = tool.inputs
    input_json = {}

    inputs_dir = os.path.join(local_working_directory, "_inputs")

    def simple_value(input, param_dict_value, cwl_type=None):
        # Hmm... cwl_type isn't really the cwl type in every case,
        # like in the case of json for instance.
        if cwl_type is None:
            input_type = input.type
            cwl_type = GALAXY_TO_CWL_TYPES[input_type]

        if cwl_type == "null":
            assert param_dict_value is None
            return None
        if cwl_type == "File":
            dataset_wrapper = param_dict_value
            extra_files_path = dataset_wrapper.extra_files_path
            secondary_files_path = os.path.join(extra_files_path, "__secondary_files__")
            path = str(dataset_wrapper)
            if os.path.exists(secondary_files_path):
                safe_makedirs(inputs_dir)
                name = os.path.basename(path)
                new_input_path = os.path.join(inputs_dir, name)
                os.symlink(path, new_input_path)
                for secondary_file_name in os.listdir(secondary_files_path):
                    secondary_file_path = os.path.join(secondary_files_path, secondary_file_name)
                    os.symlink(secondary_file_path, new_input_path + secondary_file_name)
                path = new_input_path

            return {"location": path,
                    "class": "File"}
        elif cwl_type == "integer":
            return int(str(param_dict_value))
        elif cwl_type == "long":
            return int(str(param_dict_value))
        elif cwl_type == "float":
            return float(str(param_dict_value))
        elif cwl_type == "double":
            return float(str(param_dict_value))
        elif cwl_type == "boolean":
            return string_as_bool(param_dict_value)
        elif cwl_type == "text":
            return str(param_dict_value)
        elif cwl_type == "json":
            raw_value = param_dict_value.value
            log.info("raw_value is %s (%s)" % (raw_value, type(raw_value)))
            return json.loads(raw_value)
        else:
            return str(param_dict_value)

    for input_name, input in inputs.iteritems():
        if input.type == "repeat":
            only_input = input.inputs.values()[0]
            array_value = []
            for instance in param_dict[input_name]:
                array_value.append(simple_value(only_input, instance[input_name[:-len("_repeat")]]))
            input_json[input_name[:-len("_repeat")]] = array_value
        elif input.type == "conditional":
            assert input_name in param_dict, "No value for %s in %s" % (input_name, param_dict)
            current_case = param_dict[input_name]["_cwl__type_"]
            if str(current_case) != "null":  # str because it is a wrapped...
                case_index = input.get_current_case( current_case )
                case_input = input.cases[ case_index ].inputs["_cwl__value_"]
                case_value = param_dict[input_name]["_cwl__value_"]
                input_json[input_name] = simple_value(case_input, case_value, cwl_type=current_case)
        else:
            input_json[input_name] = simple_value(input, param_dict[input_name])

    return input_json


def to_galaxy_parameters(tool, as_dict):
    """ Tool is Galaxy's representation of the tool and as_dict is a Galaxified
    representation of the input json (no paths, HDA references for instance).
    """
    inputs = tool.inputs
    galaxy_request = {}

    def from_simple_value(input, param_dict_value, cwl_type=None):
        if cwl_type == "json":
            return json.dumps(param_dict_value)
        else:
            return param_dict_value

    for input_name, input in inputs.iteritems():
        as_dict_value = as_dict.get(input_name, NOT_PRESENT)
        galaxy_input_type = input.type

        if galaxy_input_type == "repeat":
            if input_name not in as_dict:
                continue

            only_input = input.inputs.values()[0]
            for index, value in enumerate(as_dict_value):
                key = "%s_repeat_0|%s" % (input_name, only_input.name)
                galaxy_value = from_simple_value(only_input, value)
                galaxy_request[key] = galaxy_value
        elif galaxy_input_type == "conditional":
            case_strings = input.case_strings
            # TODO: less crazy handling of defaults...
            if (as_dict_value is NOT_PRESENT or as_dict_value is None) and "null" in case_strings:
                cwl_type = "null"
            elif (as_dict_value is NOT_PRESENT or as_dict_value is None):
                raise RequestParameterInvalidException(
                    "Cannot translate CWL datatype - value [%s] of type [%s] with case_strings [%s]. Non-null property must be set." % (
                        as_dict_value, type(as_dict_value), case_strings
                    )
                )
            elif isinstance(as_dict_value, bool) and "boolean" in case_strings:
                cwl_type = "boolean"
            elif isinstance(as_dict_value, int) and "integer" in case_strings:
                cwl_type = "integer"
            elif isinstance(as_dict_value, int) and "long" in case_strings:
                cwl_type = "long"
            elif isinstance(as_dict_value, (int, float)) and "float" in case_strings:
                cwl_type = "float"
            elif isinstance(as_dict_value, (int, float)) and "double" in case_strings:
                cwl_type = "double"
            elif isinstance(as_dict_value, string_types) and "string" in case_strings:
                cwl_type = "string"
            elif isinstance(as_dict_value, dict) and "src" in as_dict_value and "id" in as_dict_value:
                # Bit problematic...
                cwl_type = "File"
            elif "json" in case_strings and as_dict_value is not None:
                cwl_type = "json"
            else:
                raise RequestParameterInvalidException(
                    "Cannot translate CWL datatype - value [%s] of type [%s] with case_strings [%s]." % (
                        as_dict_value, type(as_dict_value), case_strings
                    )
                )
            galaxy_request["%s|_cwl__type_" % input_name] = cwl_type
            if cwl_type != "null":
                current_case_index = input.get_current_case(cwl_type)
                current_case_inputs = input.cases[ current_case_index ].inputs
                current_case_input = current_case_inputs[ "_cwl__value_" ]
                galaxy_value = from_simple_value(current_case_input, as_dict_value, cwl_type)
                galaxy_request["%s|_cwl__value_" % input_name] = galaxy_value
        elif as_dict_value is NOT_PRESENT:
            continue
        else:
            galaxy_value = from_simple_value(input, as_dict_value)
            galaxy_request[input_name] = galaxy_value

    log.info("Converted galaxy_request is %s" % galaxy_request)
    return galaxy_request



""" This module define an abstract class for reasoning about Galaxy's
dataset collection after jobs are finished.
"""

from galaxy.util import asbool

DEFAULT_EXTRA_FILENAME_PATTERN = r"primary_DATASET_ID_(?P<designation>[^_]+)_(?P<visible>[^_]+)_(?P<ext>[^_]+)(_(?P<dbkey>[^_]+))?"
DEFAULT_SORT_BY = "filename"
DEFAULT_SORT_COMP = "lexical"


# XML can describe custom patterns, but these literals describe named
# patterns that will be replaced.
NAMED_PATTERNS = {
    "__default__": DEFAULT_EXTRA_FILENAME_PATTERN,
    "__name__": r"(?P<name>.*)",
    "__designation__": r"(?P<designation>.*)",
    "__name_and_ext__": r"(?P<name>.*)\.(?P<ext>[^\.]+)?",
    "__designation_and_ext__": r"(?P<designation>.*)\.(?P<ext>[^\._]+)?",
}

INPUT_DBKEY_TOKEN = "__input__"
LEGACY_DEFAULT_DBKEY = None  # don't use __input__ for legacy default collection


def dataset_collector_descriptions_from_elem( elem, legacy=True ):
    primary_dataset_elems = elem.findall( "discover_datasets" )
    if len(primary_dataset_elems) == 0 and legacy:
        return [ DEFAULT_DATASET_COLLECTOR_DESCRIPTION ]
    else:
        return map( lambda elem: DatasetCollectionDescription( **elem.attrib ), primary_dataset_elems )


def dataset_collector_descriptions_from_list( discover_datasets_dicts ):
    return map( lambda kwds: DatasetCollectionDescription( **kwds ), discover_datasets_dicts )


class DatasetCollectionDescription(object):

    def __init__( self, **kwargs ):
        pattern = kwargs.get( "pattern", "__default__" )
        if pattern in NAMED_PATTERNS:
            pattern = NAMED_PATTERNS.get( pattern )
        self.pattern = pattern
        self.default_dbkey = kwargs.get( "dbkey", INPUT_DBKEY_TOKEN )
        self.default_ext = kwargs.get( "ext", None )
        if self.default_ext is None and "format" in kwargs:
            self.default_ext = kwargs.get( "format" )
        self.default_visible = asbool( kwargs.get( "visible", None ) )
        self.directory = kwargs.get( "directory", None )
        self.assign_primary_output = asbool( kwargs.get( 'assign_primary_output', False ) )
        sort_by = kwargs.get( "sort_by", DEFAULT_SORT_BY )
        if sort_by.startswith("reverse_"):
            self.sort_reverse = True
            sort_by = sort_by[len("reverse_"):]
        else:
            self.sort_reverse = False
        if "_" in sort_by:
            sort_comp, sort_by = sort_by.split("_", 1)
            assert sort_comp in ["lexical", "numeric"]
        else:
            sort_comp = DEFAULT_SORT_COMP
        assert sort_by in [
            "filename",
            "name",
            "designation",
            "dbkey"
        ]
        self.sort_key = sort_by
        self.sort_comp = sort_comp


DEFAULT_DATASET_COLLECTOR_DESCRIPTION = DatasetCollectionDescription(
    default_dbkey=LEGACY_DEFAULT_DBKEY,
)



"""Provides utilities for working with image files."""
import imghdr
import logging

try:
    import Image as PIL
except ImportError:
    try:
        from PIL import Image as PIL
    except:
        PIL = None

log = logging.getLogger(__name__)


def image_type( filename ):
    fmt = None
    if PIL is not None:
        try:
            im = PIL.open( filename )
            fmt = im.format
            im.close()
        except:
            # We continue to try with imghdr, so this is a rare case of an
            # exception we expect to happen frequently, so we're not logging
            pass
    if not fmt:
        fmt = imghdr.what( filename )
    if fmt:
        return fmt.upper()
    else:
        return False


def check_image_type( filename, types ):
    fmt = image_type( filename )
    if fmt in types:
        return True
    return False


def get_image_ext( file_path ):
    # determine ext
    fmt = image_type( file_path )
    if fmt in [ 'JPG', 'JPEG' ]:
        return 'jpg'
    if fmt == 'PNG':
        return 'png'
    if fmt == 'TIFF':
        return 'tiff'
    if fmt == 'BMP':
        return 'bmp'
    if fmt == 'GIF':
        return 'gif'
    if fmt == 'IM':
        return 'im'
    if fmt == 'PCD':
        return 'pcd'
    if fmt == 'PCX':
        return 'pcx'
    if fmt == 'PPM':
        return 'ppm'
    if fmt == 'PSD':
        return 'psd'
    if fmt == 'XBM':
        return 'xbm'
    if fmt == 'XPM':
        return 'xpm'
    if fmt == 'RGB':
        return 'rgb'
    if fmt == 'PBM':
        return 'pbm'
    if fmt == 'PGM':
        return 'pgm'
    return None



from .interface import ToolLineage, ToolLineageVersion

try:
    from galaxy.model.tool_shed_install import ToolVersion
except ImportError:
    ToolVersion = None


class ToolVersionCache(object):
    """
    Instances of this class allow looking up tool_version objects from memory
    (instead of querying the database) using the tool_version id or the tool_id.
    This is used to lookup parent tool_version ids using child tool_id, or the
    inverse, and getting all previous/next tool versions without numerous
    database requests.
    """
    def __init__(self, app):
        self.app = app
        self.tool_version_by_id, self.tool_version_by_tool_id = self.get_tool_versions()
        self.tool_id_to_parent_id, self.parent_id_to_tool_id = self.get_tva_map()

    def get_tva_map(self):
        """
        Retrieves all ToolVersionAssociation objects from the database, and builds
        dictionaries that can be used to either get a tools' parent tool_version id
        (which can be used to get the parent's tool_version object), or to get the
        child's tool id using the parent's tool_version id.
        """
        tvas = self.app.install_model.context.query(self.app.install_model.ToolVersionAssociation).all()
        tool_id_to_parent_id = {tva.tool_id: tva.parent_id for tva in tvas}
        parent_id_to_tool_id = {tva.parent_id: tva.tool_id for tva in tvas}
        return tool_id_to_parent_id, parent_id_to_tool_id

    def get_tool_versions(self):
        """
        Get all tool_version objects from the database and build 2 dictionaries,
        with tool_version id or tool_id as key and the tool_version object as value.
        """
        tool_versions = self.app.install_model.context.query(self.app.install_model.ToolVersion).all()
        tool_version_by_id = {tv.id: tv for tv in tool_versions}
        tool_version_by_tool_id = {tv.tool_id: tv for tv in tool_versions}
        return tool_version_by_id, tool_version_by_tool_id


class ToolShedLineage(ToolLineage):
    """ Representation of tool lineage derived from tool shed repository
    installations. """

    def __init__(self, app, tool_version, tool_shed_repository=None):
        if ToolVersion is None:
            raise Exception("Tool shed models not present, can't create tool shed lineages.")
        self.app = app
        self.tool_version_id = tool_version.id
        # Only used for logging
        self._tool_shed_repository = tool_shed_repository

    @staticmethod
    def from_tool( app, tool ):
        # Make sure the tool has a tool_version.
        if not get_installed_tool_version( app, tool.id ):
            tool_version = ToolVersion( tool_id=tool.id, tool_shed_repository=tool.tool_shed_repository )
            app.install_model.context.add( tool_version )
            app.install_model.context.flush()
        return ToolShedLineage( app, tool.tool_version )

    @staticmethod
    def from_tool_id( app, tool_id ):
        tool_version = get_installed_tool_version( app, tool_id )
        if tool_version:
            return ToolShedLineage( app, tool_version )
        else:
            return None

    def get_version_ids( self, reverse=False ):
        tool_version = self.app.install_model.context.query( ToolVersion ).get( self.tool_version_id )
        result = tool_version.get_version_ids( self.app, reverse=reverse )
        return result

    def get_versions( self, reverse=False ):
        return [ ToolLineageVersion.from_guid(_) for _ in self.get_version_ids( reverse=reverse ) ]

    def to_dict(self):
        tool_shed_repository = self._tool_shed_repository
        rval = dict(
            tool_version_id=self.tool_version_id,
            tool_versions=[v.to_dict() for v in self.get_versions()],
            tool_shed_repository=tool_shed_repository if tool_shed_repository is not None else None,
            lineage_type='tool_shed',
        )
        return rval


def get_installed_tool_version( app, tool_id ):
    return app.tool_version_cache.tool_version_by_tool_id.get(tool_id, None)


__all__ = ( "ToolShedLineage", )



""" This module provides proxy objects around objects from the common
workflow language reference implementation library cwltool. These proxies
adapt cwltool to Galaxy features and abstract the library away from the rest
of the framework.
"""
from __future__ import absolute_import

import json
import logging
import os
from abc import ABCMeta, abstractmethod

import six

from galaxy.util import safe_makedirs
from galaxy.util.bunch import Bunch
from galaxy.util.odict import odict

from .cwltool_deps import (
    ensure_cwltool_available,
    process,
)

from .schema import non_strict_schema_loader, schema_loader

log = logging.getLogger(__name__)

JOB_JSON_FILE = ".cwl_job.json"
SECONDARY_FILES_EXTRA_PREFIX = "__secondary_files__"


SUPPORTED_TOOL_REQUIREMENTS = [
    "CreateFileRequirement",
    "DockerRequirement",
    "EnvVarRequirement",
    "InlineJavascriptRequirement",
]


SUPPORTED_WORKFLOW_REQUIREMENTS = SUPPORTED_TOOL_REQUIREMENTS + [
]


def tool_proxy(tool_path, strict_cwl_validation=True):
    """ Provide a proxy object to cwltool data structures to just
    grab relevant data.
    """
    ensure_cwltool_available()
    tool = to_cwl_tool_object(tool_path, strict_cwl_validation=strict_cwl_validation)
    return tool


def workflow_proxy(workflow_path, strict_cwl_validation=True):
    ensure_cwltool_available()
    workflow = to_cwl_workflow_object(workflow_path, strict_cwl_validation=strict_cwl_validation)
    return workflow


def load_job_proxy(job_directory, strict_cwl_validation=True):
    ensure_cwltool_available()
    job_objects_path = os.path.join(job_directory, JOB_JSON_FILE)
    job_objects = json.load(open(job_objects_path, "r"))
    tool_path = job_objects["tool_path"]
    job_inputs = job_objects["job_inputs"]
    output_dict = job_objects["output_dict"]
    cwl_tool = tool_proxy(tool_path, strict_cwl_validation=strict_cwl_validation)
    cwl_job = cwl_tool.job_proxy(job_inputs, output_dict, job_directory=job_directory)
    return cwl_job


def to_cwl_tool_object(tool_path, strict_cwl_validation=True):
    proxy_class = None
    cwl_tool = _schema_loader(strict_cwl_validation).tool(path=tool_path)
    if isinstance(cwl_tool, int):
        raise Exception("Failed to load tool.")

    raw_tool = cwl_tool.tool
    check_requirements(raw_tool)
    if "class" not in raw_tool:
        raise Exception("File does not declare a class, not a valid Draft 3+ CWL tool.")
    process_class = raw_tool["class"]
    if process_class == "CommandLineTool":
        proxy_class = CommandLineToolProxy
    elif process_class == "ExpressionTool":
        proxy_class = ExpressionToolProxy
    else:
        raise Exception("File not a CWL CommandLineTool.")
    if "cwlVersion" not in raw_tool:
        raise Exception("File does not declare a CWL version, pre-draft 3 CWL tools are not supported.")

    proxy = proxy_class(cwl_tool, tool_path)
    return proxy


def to_cwl_workflow_object(workflow_path, strict_cwl_validation=None):
    proxy_class = WorkflowProxy
    cwl_workflow = _schema_loader(strict_cwl_validation).tool(path=workflow_path)
    raw_workflow = cwl_workflow.tool
    check_requirements(raw_workflow, tool=False)

    proxy = proxy_class(cwl_workflow, workflow_path)
    return proxy


def _schema_loader(strict_cwl_validation):
    target_schema_loader = schema_loader if strict_cwl_validation else non_strict_schema_loader
    return target_schema_loader


def check_requirements(rec, tool=True):
    if isinstance(rec, dict):
        if "requirements" in rec:
            for r in rec["requirements"]:
                if tool:
                    possible = SUPPORTED_TOOL_REQUIREMENTS
                else:
                    possible = SUPPORTED_WORKFLOW_REQUIREMENTS
                if r["class"] not in possible:
                    raise Exception("Unsupported requirement %s" % r["class"])
        for d in rec:
            check_requirements(rec[d], tool=tool)
    if isinstance(rec, list):
        for d in rec:
            check_requirements(d, tool=tool)


@six.add_metaclass(ABCMeta)
class ToolProxy( object ):

    def __init__(self, tool, tool_path):
        self._tool = tool
        self._tool_path = tool_path

    def job_proxy(self, input_dict, output_dict, job_directory="."):
        """ Build a cwltool.job.Job describing computation using a input_json
        Galaxy will generate mapping the Galaxy description of the inputs into
        a cwltool compatible variant.
        """
        return JobProxy(self, input_dict, output_dict, job_directory=job_directory)

    @abstractmethod
    def input_instances(self):
        """ Return InputInstance objects describing mapping to Galaxy inputs. """

    @abstractmethod
    def output_instances(self):
        """ Return OutputInstance objects describing mapping to Galaxy inputs. """

    @abstractmethod
    def docker_identifier(self):
        """ Return docker identifier for embedding in tool description. """

    @abstractmethod
    def description(self):
        """ Return description to tool. """

    @abstractmethod
    def label(self):
        """ Return label for tool. """


class CommandLineToolProxy(ToolProxy):

    def description(self):
        return self._tool.tool.get('doc')

    def label(self):
        return self._tool.tool.get('label')

    def input_instances(self):
        return self._find_inputs(self._tool.inputs_record_schema)

    def _find_inputs(self, schema):
        schema_type = schema["type"]
        if isinstance(schema_type, list):
            raise Exception("Union types not yet implemented.")
        elif isinstance(schema_type, dict):
            return self._find_inputs(schema_type)
        else:
            if schema_type in self._tool.schemaDefs:
                schema = self._tool.schemaDefs[schema_type]

            if schema["type"] == "record":
                return [_simple_field_to_input(_) for _ in schema["fields"]]

    def output_instances(self):
        outputs_schema = self._tool.outputs_record_schema
        return self._find_outputs(outputs_schema)

    def _find_outputs(self, schema):
        rval = []
        if not rval and schema["type"] == "record":
            for output in schema["fields"]:
                # output_type = output.get("type", None)
                # if output_type != "File":
                #     template = "Unhandled output type [%s] encountered."
                #     raise Exception(template % output_type)
                rval.append(_simple_field_to_output(output))

        return rval

    def docker_identifier(self):
        tool = self._tool.tool
        reqs_and_hints = tool.get("requirements", []) + tool.get("hints", [])
        for hint in reqs_and_hints:
            if hint["class"] == "DockerRequirement":
                if "dockerImageId" in hint:
                    return hint["dockerImageId"]
                else:
                    return hint["dockerPull"]
        return None


class ExpressionToolProxy(CommandLineToolProxy):
    pass


class JobProxy(object):

    def __init__(self, tool_proxy, input_dict, output_dict, job_directory):
        self._tool_proxy = tool_proxy
        self._input_dict = input_dict
        self._output_dict = output_dict
        self._job_directory = job_directory

        self._final_output = []
        self._ok = True
        self._cwl_job = None
        self._is_command_line_job = None

    def cwl_job(self):
        self._ensure_cwl_job_initialized()
        return self._cwl_job

    @property
    def is_command_line_job(self):
        self._ensure_cwl_job_initialized()
        assert self._is_command_line_job is not None
        return self._is_command_line_job

    def _ensure_cwl_job_initialized(self):
        if self._cwl_job is None:

            self._cwl_job = next(self._tool_proxy._tool.job(
                self._input_dict,
                self._output_callback,
                basedir=self._job_directory,
                select_resources=self._select_resources,
                use_container=False
            ))
            self._is_command_line_job = hasattr(self._cwl_job, "command_line")

    def _select_resources(self, request):
        new_request = request.copy()
        new_request["cores"] = "$GALAXY_SLOTS"
        return new_request

    @property
    def command_line(self):
        if self.is_command_line_job:
            return self.cwl_job().command_line
        else:
            return ["true"]

    @property
    def stdin(self):
        if self.is_command_line_job:
            return self.cwl_job().stdin
        else:
            return None

    @property
    def stdout(self):
        if self.is_command_line_job:
            return self.cwl_job().stdout
        else:
            return None

    @property
    def environment(self):
        if self.is_command_line_job:
            return self.cwl_job().environment
        else:
            return {}

    @property
    def generate_files(self):
        if self.is_command_line_job:
            return self.cwl_job().generatefiles
        else:
            return {}

    def _output_callback(self, out, process_status):
        if process_status == "success":
            self._final_output = out
        else:
            self._ok = False

        log.info("Output are %s, status is %s" % (out, process_status))

    def collect_outputs(self, tool_working_directory):
        if not self.is_command_line_job:
            self.cwl_job().run(
            )
            return self._final_output
        else:
            return self.cwl_job().collect_outputs(tool_working_directory)

    def save_job(self):
        job_file = JobProxy._job_file(self._job_directory)
        job_objects = {
            "tool_path": os.path.abspath(self._tool_proxy._tool_path),
            "job_inputs": self._input_dict,
            "output_dict": self._output_dict,
        }
        json.dump(job_objects, open(job_file, "w"))

    def _output_extra_files_dir(self, output_name):
        output_id = self.output_id(output_name)
        return os.path.join(self._job_directory, "dataset_%s_files" % output_id)

    def output_id(self, output_name):
        output_id = self._output_dict[output_name]["id"]
        return output_id

    def output_path(self, output_name):
        output_id = self._output_dict[output_name]["path"]
        return output_id

    def output_secondary_files_dir(self, output_name, create=False):
        extra_files_dir = self._output_extra_files_dir(output_name)
        secondary_files_dir = os.path.join(extra_files_dir, SECONDARY_FILES_EXTRA_PREFIX)
        if create and not os.path.exists(secondary_files_dir):
            safe_makedirs(secondary_files_dir)
        return secondary_files_dir

    def stage_files(self):
        cwl_job = self.cwl_job()
        if hasattr(cwl_job, "pathmapper"):
            process.stageFiles(self.cwl_job().pathmapper, os.symlink, ignoreWritable=True)
        # else: expression tools do not have a path mapper.

    @staticmethod
    def _job_file(job_directory):
        return os.path.join(job_directory, JOB_JSON_FILE)


class WorkflowProxy(object):

    def __init__(self, workflow, workflow_path):
        self._workflow = workflow
        self._workflow_path = workflow_path

    def step_proxies(self):
        proxies = []
        for step in self._workflow.steps:
            proxies.append(StepProxy(self, step))
        return proxies

    @property
    def runnables(self):
        runnables = []
        for step in self._workflow.steps:
            if "run" in step.tool:
                runnables.append(step.tool["run"])
        return runnables

    def to_dict(self):
        name = os.path.basename(self._workflow_path)
        steps = {}

        index = 0
        for i, input_dict in self._workflow.tool['inputs']:
            index += 1
            steps[index] = input_dict

        for i, step_proxy in enumerate(self.step_proxies()):
            index += 1
            steps[index] = step_proxy.to_dict()

        return {
            'name': name,
            'steps': steps,
        }


class StepProxy(object):

    def __init__(self, workflow_proxy, step):
        self._workflow_proxy = workflow_proxy
        self._step = step

    def to_dict(self):
        return {}


def _simple_field_union(field):
    field_type = _field_to_field_type(field)  # Must be a list if in here?

    def any_of_in_field_type(types):
        return any([t in field_type for t in types])

    name, label, description = _field_metadata(field)

    case_name = "_cwl__type_"
    case_label = "Specify Parameter %s As" % label

    def value_input(**kwds):
        value_name = "_cwl__value_"
        value_label = label
        value_description = description
        return InputInstance(
            value_name,
            value_label,
            value_description,
            **kwds
        )

    select_options = []
    case_options = []
    if "null" in field_type:
        select_options.append({"value": "null", "label": "None", "selected": True})
        case_options.append(("null", []))
    if any_of_in_field_type(["Any", "string"]):
        select_options.append({"value": "string", "label": "Simple String"})
        case_options.append(("string", [value_input(input_type=INPUT_TYPE.TEXT)]))
    if any_of_in_field_type(["Any", "boolean"]):
        select_options.append({"value": "boolean", "label": "Boolean"})
        case_options.append(("boolean", [value_input(input_type=INPUT_TYPE.BOOLEAN)]))
    if any_of_in_field_type(["Any", "int"]):
        select_options.append({"value": "int", "label": "Integer"})
        case_options.append(("int", [value_input(input_type=INPUT_TYPE.INTEGER)]))
    if any_of_in_field_type(["Any", "float"]):
        select_options.append({"value": "float", "label": "Floating Point Number"})
        case_options.append(("float", [value_input(input_type=INPUT_TYPE.FLOAT)]))
    if any_of_in_field_type(["Any", "File"]):
        select_options.append({"value": "data", "label": "Dataset"})
        case_options.append(("data", [value_input(input_type=INPUT_TYPE.DATA)]))
    if "Any" in field_type:
        select_options.append({"value": "json", "label": "JSON Data Structure"})
        case_options.append(("json", [value_input(input_type=INPUT_TYPE.TEXT, area=True)]))

    case_input = SelectInputInstance(
        name=case_name,
        label=case_label,
        description=False,
        options=select_options,
    )

    return ConditionalInstance(name, case_input, case_options)


def _simple_field_to_input(field):
    field_type = _field_to_field_type(field)
    if isinstance(field_type, list):
        # Length must be greater than 1...
        return _simple_field_union(field)

    name, label, description = _field_metadata(field)

    type_kwds = _simple_field_to_input_type_kwds(field)
    return InputInstance(name, label, description, **type_kwds)


def _simple_field_to_input_type_kwds(field, field_type=None):
    simple_map_type_map = {
        "File": INPUT_TYPE.DATA,
        "int": INPUT_TYPE.INTEGER,
        "long": INPUT_TYPE.INTEGER,
        "float": INPUT_TYPE.INTEGER,
        "double": INPUT_TYPE.INTEGER,
        "string": INPUT_TYPE.TEXT,
        "boolean": INPUT_TYPE.BOOLEAN,
    }

    if field_type is None:
        field_type = _field_to_field_type(field)

    if field_type in simple_map_type_map.keys():
        input_type = simple_map_type_map[field_type]
        return {"input_type": input_type, "array": False}
    elif field_type == "array":
        if isinstance(field["type"], dict):
            array_type = field["type"]["items"]
        else:
            array_type = field["items"]
        if array_type in simple_map_type_map.keys():
            input_type = simple_map_type_map[array_type]
        return {"input_type": input_type, "array": True}
    else:
        raise Exception("Unhandled simple field type encountered - [%s]." % field_type)


def _field_to_field_type(field):
    field_type = field["type"]
    if isinstance(field_type, dict):
        field_type = field_type["type"]
    if isinstance(field_type, list):
        field_type_length = len(field_type)
        if field_type_length == 0:
            raise Exception("Zero-length type list encountered, invalid CWL?")
        elif len(field_type) == 1:
            field_type = field_type[0]

    if field_type == "Any":
        field_type = ["Any"]

    return field_type


def _field_metadata(field):
    name = field["name"]
    label = field.get("label", None)
    description = field.get("doc", None)
    return name, label, description


def _simple_field_to_output(field):
    name = field["name"]
    output_data_class = field["type"]
    output_instance = OutputInstance(
        name,
        output_data_type=output_data_class,
        output_type=OUTPUT_TYPE.GLOB
    )
    return output_instance


INPUT_TYPE = Bunch(
    DATA="data",
    INTEGER="integer",
    FLOAT="float",
    TEXT="text",
    BOOLEAN="boolean",
    SELECT="select",
    CONDITIONAL="conditional",
)


class ConditionalInstance(object):

    def __init__(self, name, case, whens):
        self.input_type = INPUT_TYPE.CONDITIONAL
        self.name = name
        self.case = case
        self.whens = whens

    def to_dict(self):

        as_dict = dict(
            name=self.name,
            type=INPUT_TYPE.CONDITIONAL,
            test=self.case.to_dict(),
            when=odict(),
        )
        for value, block in self.whens:
            as_dict["when"][value] = [i.to_dict() for i in block]

        return as_dict


class SelectInputInstance(object):

    def __init__(self, name, label, description, options):
        self.input_type = INPUT_TYPE.SELECT
        self.name = name
        self.label = label
        self.description = description
        self.options = options

    def to_dict(self):
        # TODO: serialize options...
        as_dict = dict(
            name=self.name,
            label=self.label or self.name,
            help=self.description,
            type=self.input_type,
            options=self.options,
        )
        return as_dict


class InputInstance(object):

    def __init__(self, name, label, description, input_type, array=False, area=False):
        self.input_type = input_type
        self.name = name
        self.label = label
        self.description = description
        self.required = True
        self.array = array
        self.area = area

    def to_dict(self, itemwise=True):
        if itemwise and self.array:
            as_dict = dict(
                type="repeat",
                name="%s_repeat" % self.name,
                title="%s" % self.name,
                blocks=[
                    self.to_dict(itemwise=False)
                ]
            )
        else:
            as_dict = dict(
                name=self.name,
                label=self.label or self.name,
                help=self.description,
                type=self.input_type,
                optional=not self.required,
            )
            if self.area:
                as_dict["area"] = True

            if self.input_type == INPUT_TYPE.INTEGER:
                as_dict["value"] = "0"
            if self.input_type == INPUT_TYPE.FLOAT:
                as_dict["value"] = "0.0"
        return as_dict


OUTPUT_TYPE = Bunch(
    GLOB="glob",
    STDOUT="stdout",
)


class OutputInstance(object):

    def __init__(self, name, output_data_type, output_type, path=None):
        self.name = name
        self.output_data_type = output_data_type
        self.output_type = output_type
        self.path = path


__all__ = (
    'tool_proxy',
    'load_job_proxy',
)



from .stock import StockLineage
from .tool_shed import ToolShedLineage


def remove_version_from_guid( guid ):
    """
    Removes version from toolshed-derived tool_id(=guid).
    """
    if "/repos/" not in guid:
        return None
    last_slash = guid.rfind('/')
    return guid[:last_slash]


class LineageMap(object):
    """ Map each unique tool id to a lineage object.
    """

    def __init__(self, app):
        self.lineage_map = {}
        self.app = app

    def register(self, tool, from_toolshed=False):
        tool_id = tool.id
        versionless_tool_id = remove_version_from_guid( tool_id )
        if from_toolshed:
            lineage = ToolShedLineage.from_tool(self.app, tool)
        else:
            lineage = StockLineage.from_tool( tool )
        if versionless_tool_id and versionless_tool_id not in self.lineage_map:
            self.lineage_map[versionless_tool_id] = lineage
        if tool_id not in self.lineage_map:
            self.lineage_map[tool_id] = lineage
        return self.lineage_map[tool_id]

    def get(self, tool_id):
        if tool_id not in self.lineage_map:
            lineage = ToolShedLineage.from_tool_id( self.app, tool_id )
            if lineage:
                self.lineage_map[tool_id] = lineage

        return self.lineage_map.get(tool_id, None)

    def get_versionless(self, tool_id):
        versionless_tool_id = remove_version_from_guid(tool_id)
        return self.lineage_map.get(versionless_tool_id, None)


__all__ = ("LineageMap", )



"""Utilities for working with mulled abstractions outside the mulled package."""
from __future__ import print_function

import collections
import hashlib

from distutils.version import LooseVersion

try:
    import requests
except ImportError:
    requests = None


def create_repository(namespace, repo_name, oauth_token):
    assert oauth_token
    headers = {'Authorization': 'Bearer %s' % oauth_token}
    data = {
        "repository": repo_name,
        "namespace": namespace,
        "description": "",
        "visibility": "public",
    }
    requests.post("https://quay.io/api/v1/repository", json=data, headers=headers)


def quay_versions(namespace, pkg_name):
    """Get all version tags for a Docker image stored on quay.io for supplied package name."""
    data = quay_repository(namespace, pkg_name)

    if 'error_type' in data and data['error_type'] == "invalid_token":
        return []

    if 'tags' not in data:
        raise Exception("Unexpected response from quay.io - not tags description found [%s]" % data)

    return [tag for tag in data['tags'] if tag != 'latest']


def quay_repository(namespace, pkg_name):
    if requests is None:
        raise Exception("requets library is unavailable, functionality not available.")

    assert namespace is not None
    assert pkg_name is not None
    url = 'https://quay.io/api/v1/repository/%s/%s' % (namespace, pkg_name)
    response = requests.get(url, timeout=None)
    data = response.json()
    return data


def mulled_tags_for(namespace, image):
    """Fetch remote tags available for supplied image name.

    The result will be sorted so newest tags are first.
    """
    tags = quay_versions(namespace, image)
    tags = version_sorted(tags)
    return tags


def split_tag(tag):
    """Split mulled image name into conda version and conda build."""
    version = tag.split('--', 1)[0]
    build = tag.split('--', 1)[1]
    return version, build


def version_sorted(elements):
    """Sort iterable based on loose description of "version" from newest to oldest."""
    return sorted(elements, key=LooseVersion, reverse=True)


Target = collections.namedtuple("Target", ["package_name", "version", "build"])


def build_target(package_name, version=None, build=None, tag=None):
    """Use supplied arguments to build a :class:`Target` object."""
    if tag is not None:
        assert version is None
        assert build is None
        version, build = split_tag(tag)

    return Target(package_name, version, build)


def conda_build_target_str(target):
    rval = target.package_name
    if target.version:
        rval += "=%s" % target.version

        if target.build:
            rval += "=%s" % target.build

    return rval


def image_name(targets, image_build=None, name_override=None):
    if name_override is not None:
        print("WARNING: Overriding mulled image name, auto-detection of 'mulled' package attributes will fail to detect result.")
        return name_override

    targets = list(targets)
    if len(targets) == 1:
        target = targets[0]
        suffix = ""
        if target.version is not None:
            if image_build is not None:
                print("WARNING: Hard-coding image build instead of using Conda build - this is not recommended.")
                suffix = image_build
            else:
                suffix += ":%s" % target.version
                build = target.build
                if build is not None:
                    suffix += "--%s" % build
        return "%s%s" % (target.package_name, suffix)
    else:
        targets_order = sorted(targets, key=lambda t: t.package_name)
        requirements_buffer = "\n".join(map(conda_build_target_str, targets_order))
        m = hashlib.sha1()
        m.update(requirements_buffer.encode())
        suffix = "" if not image_build else ":%s" % image_build
        return "mulled-v1-%s%s" % (m.hexdigest(), suffix)


__all__ = (
    "build_target",
    "conda_build_target_str",
    "image_name",
    "mulled_tags_for",
    "quay_versions",
    "split_tag",
    "Target",
    "version_sorted",
)



"""
This is still an experimental module and there will almost certainly be backward
incompatible changes coming.
"""

import logging
import os
import re

import galaxy.tools.deps.installable
import galaxy.tools.deps.requirements

from ..conda_util import (
    build_isolated_environment,
    cleanup_failed_install,
    cleanup_failed_install_of_environment,
    CondaContext,
    CondaTarget,
    hash_conda_packages,
    install_conda,
    install_conda_target,
    install_conda_targets,
    installed_conda_targets,
    is_conda_target_installed,
    USE_PATH_EXEC_DEFAULT,
)
from ..resolvers import (
    Dependency,
    DependencyException,
    DependencyResolver,
    InstallableDependencyResolver,
    ListableDependencyResolver,
    MappableDependencyResolver,
    MultipleDependencyResolver,
    NullDependency,
    SpecificationPatternDependencyResolver,
)


DEFAULT_BASE_PATH_DIRECTORY = "_conda"
DEFAULT_CONDARC_OVERRIDE = "_condarc"
DEFAULT_ENSURE_CHANNELS = "iuc,bioconda,r,defaults,conda-forge"

log = logging.getLogger(__name__)


class CondaDependencyResolver(DependencyResolver, MultipleDependencyResolver, ListableDependencyResolver, InstallableDependencyResolver, SpecificationPatternDependencyResolver, MappableDependencyResolver):
    dict_collection_visible_keys = DependencyResolver.dict_collection_visible_keys + ['conda_prefix', 'versionless', 'ensure_channels', 'auto_install']
    resolver_type = "conda"
    config_options = {
        'prefix': None,
        'exec': None,
        'debug': None,
        'ensure_channels': DEFAULT_ENSURE_CHANNELS,
        'auto_install': False,
        'auto_init': True,
        'copy_dependencies': False,
        'use_local': False,
    }
    _specification_pattern = re.compile(r"https\:\/\/anaconda.org\/\w+\/\w+")

    def __init__(self, dependency_manager, **kwds):
        self._setup_mapping(dependency_manager, **kwds)
        self.versionless = _string_as_bool(kwds.get('versionless', 'false'))
        self.dependency_manager = dependency_manager

        def get_option(name):
            return dependency_manager.get_resolver_option(self, name, explicit_resolver_options=kwds)

        # Conda context options (these define the environment)
        conda_prefix = get_option("prefix")
        if conda_prefix is None:
            conda_prefix = os.path.join(
                dependency_manager.default_base_path, DEFAULT_BASE_PATH_DIRECTORY
            )
        conda_prefix = os.path.abspath(conda_prefix)

        self.conda_prefix_parent = os.path.dirname(conda_prefix)

        condarc_override = get_option("condarc_override")
        if condarc_override is None:
            condarc_override = os.path.join(
                dependency_manager.default_base_path, DEFAULT_CONDARC_OVERRIDE
            )

        copy_dependencies = _string_as_bool(get_option("copy_dependencies"))
        use_local = _string_as_bool(get_option("use_local"))
        conda_exec = get_option("exec")
        debug = _string_as_bool(get_option("debug"))
        ensure_channels = get_option("ensure_channels")
        use_path_exec = get_option("use_path_exec")
        if use_path_exec is None:
            use_path_exec = USE_PATH_EXEC_DEFAULT
        else:
            use_path_exec = _string_as_bool(use_path_exec)
        if ensure_channels is None:
            ensure_channels = DEFAULT_ENSURE_CHANNELS

        conda_context = CondaContext(
            conda_prefix=conda_prefix,
            conda_exec=conda_exec,
            debug=debug,
            ensure_channels=ensure_channels,
            condarc_override=condarc_override,
            use_path_exec=use_path_exec,
            copy_dependencies=copy_dependencies,
            use_local=use_local,
        )
        self.ensure_channels = ensure_channels

        # Conda operations options (these define how resolution will occur)
        auto_install = _string_as_bool(get_option("auto_install"))
        self.auto_init = _string_as_bool(get_option("auto_init"))
        self.conda_context = conda_context
        self.disabled = not galaxy.tools.deps.installable.ensure_installed(conda_context, install_conda, self.auto_init)
        if self.auto_init and not self.disabled:
            self.conda_context.ensure_conda_build_installed_if_needed()
        self.auto_install = auto_install
        self.copy_dependencies = copy_dependencies

    def clean(self, **kwds):
        return self.conda_context.exec_clean()

    def install_all(self, conda_targets):
        env = self.merged_environment_name(conda_targets)
        return_code = install_conda_targets(conda_targets, env, conda_context=self.conda_context)
        if return_code != 0:
            is_installed = False
        else:
            # Recheck if installed
            is_installed = self.conda_context.has_env(env)

        if not is_installed:
            log.debug("Removing failed conda install of {}".format(str(conda_targets)))
            cleanup_failed_install_of_environment(env, conda_context=self.conda_context)

        return is_installed

    def resolve_all(self, requirements, **kwds):
        if len(requirements) == 0:
            return False

        if not os.path.isdir(self.conda_context.conda_prefix):
            return False

        for requirement in requirements:
            if requirement.type != "package":
                return False

        ToolRequirements = galaxy.tools.deps.requirements.ToolRequirements
        expanded_requirements = ToolRequirements([self._expand_requirement(r) for r in requirements])
        if self.versionless:
            conda_targets = [CondaTarget(r.name, version=None) for r in expanded_requirements]
        else:
            conda_targets = [CondaTarget(r.name, version=r.version) for r in expanded_requirements]

        preserve_python_environment = kwds.get("preserve_python_environment", False)

        env = self.merged_environment_name(conda_targets)
        dependencies = []

        is_installed = self.conda_context.has_env(env)
        install = kwds.get('install', None)
        if install is None:
            # Default behavior, install dependencies if conda_auto_install is active.
            install = not is_installed and self.auto_install
        elif install:
            # Install has been set to True, install if not yet installed.
            install = not is_installed
        if install:
            is_installed = self.install_all(conda_targets)

        if is_installed:
            for requirement in requirements:
                dependency = MergedCondaDependency(
                    self.conda_context,
                    self.conda_context.env_path(env),
                    exact=not self.versionless or requirement.version is None,
                    name=requirement.name,
                    version=requirement.version,
                    preserve_python_environment=preserve_python_environment,
                )
                dependencies.append(dependency)

        return dependencies

    def merged_environment_name(self, conda_targets):
        if len(conda_targets) > 1:
            # For continuity with mulled containers this is kind of nice.
            return "mulled-v1-%s" % hash_conda_packages(conda_targets)
        else:
            assert len(conda_targets) == 1
            return conda_targets[0].install_environment

    def resolve(self, requirement, **kwds):
        requirement = self._expand_requirement(requirement)
        name, version, type = requirement.name, requirement.version, requirement.type

        # Check for conda just not being there, this way we can enable
        # conda by default and just do nothing in not configured.
        if not os.path.isdir(self.conda_context.conda_prefix):
            return NullDependency(version=version, name=name)

        if type != "package":
            return NullDependency(version=version, name=name)

        exact = not self.versionless or version is None
        if self.versionless:
            version = None

        conda_target = CondaTarget(name, version=version)
        is_installed = is_conda_target_installed(
            conda_target, conda_context=self.conda_context
        )

        preserve_python_environment = kwds.get("preserve_python_environment", False)

        job_directory = kwds.get("job_directory", None)
        install = kwds.get('install', None)
        if install is None:
            install = not is_installed and self.auto_install
        elif install:
            install = not is_installed
        if install:
            is_installed = self.install_dependency(name=name, version=version, type=type)

        if not is_installed:
            return NullDependency(version=version, name=name)

        # Have installed conda_target and job_directory to send it to.
        # If dependency is for metadata generation, store environment in conda-metadata-env
        if kwds.get("metadata", False):
            conda_env = "conda-metadata-env"
        else:
            conda_env = "conda-env"

        if job_directory:
            conda_environment = os.path.join(job_directory, conda_env)
        else:
            conda_environment = self.conda_context.env_path(conda_target.install_environment)

        return CondaDependency(
            self.conda_context,
            conda_environment,
            exact,
            name,
            version,
            preserve_python_environment=preserve_python_environment,
        )

    def _expand_requirement(self, requirement):
        return self._expand_specs(self._expand_mappings(requirement))

    def list_dependencies(self):
        for install_target in installed_conda_targets(self.conda_context):
            name = install_target.package
            version = install_target.version
            yield self._to_requirement(name, version)

    def install_dependency(self, name, version, type, **kwds):
        "Returns True on (seemingly) successfull installation"
        if type != "package":
            log.warning("Cannot install dependencies of type '%s'" % type)
            return False

        if self.versionless:
            version = None

        conda_target = CondaTarget(name, version=version)

        is_installed = is_conda_target_installed(
            conda_target, conda_context=self.conda_context
        )

        if is_installed:
            return is_installed

        return_code = install_conda_target(conda_target, conda_context=self.conda_context)
        if return_code != 0:
            is_installed = False
        else:
            # Recheck if installed
            is_installed = is_conda_target_installed(
                conda_target, conda_context=self.conda_context
            )
        if not is_installed:
            log.debug("Removing failed conda install of {}, version '{}'".format(name, version))
            cleanup_failed_install(conda_target, conda_context=self.conda_context)

        return is_installed

    @property
    def prefix(self):
        return self.conda_context.conda_prefix


class MergedCondaDependency(Dependency):
    dict_collection_visible_keys = Dependency.dict_collection_visible_keys + ['environment_path', 'name', 'version']
    dependency_type = 'conda'

    def __init__(self, conda_context, environment_path, exact, name=None, version=None, preserve_python_environment=False):
        self.activate = conda_context.activate
        self.conda_context = conda_context
        self.environment_path = environment_path
        self._exact = exact
        self._name = name
        self._version = version
        self.cache_path = None
        self._preserve_python_environment = preserve_python_environment

    @property
    def exact(self):
        return self._exact

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    def shell_commands(self, requirement):
        if self._preserve_python_environment:
            # On explicit testing the only such requirement I am aware of is samtools - and it seems to work
            # fine with just appending the PATH as done below. Other tools may require additional
            # variables in the future.
            return """export PATH=$PATH:'%s/bin' """ % (
                self.environment_path,
            )
        else:
            return """[ "$CONDA_DEFAULT_ENV" = "%s" ] || . %s '%s' > conda_activate.log 2>&1 """ % (
                self.environment_path,
                self.activate,
                self.environment_path
            )


class CondaDependency(Dependency):
    dict_collection_visible_keys = Dependency.dict_collection_visible_keys + ['environment_path', 'name', 'version']
    dependency_type = 'conda'
    cacheable = True

    def __init__(self, conda_context, environment_path, exact, name=None, version=None, preserve_python_environment=False):
        self.activate = conda_context.activate
        self.conda_context = conda_context
        self.environment_path = environment_path
        self._exact = exact
        self._name = name
        self._version = version
        self.cache_path = None
        self._preserve_python_environment = preserve_python_environment

    @property
    def exact(self):
        return self._exact

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    def build_cache(self, cache_path):
        self.set_cache_path(cache_path)
        self.build_environment()

    def set_cache_path(self, cache_path):
        self.cache_path = cache_path
        self.environment_path = cache_path

    def build_environment(self):
        env_path, exit_code = build_isolated_environment(
            CondaTarget(self.name, self.version),
            path=self.environment_path,
            copy=self.conda_context.copy_dependencies,
            conda_context=self.conda_context,
        )
        if exit_code:
            if len(os.path.abspath(self.environment_path)) > 79:
                # TODO: remove this once conda_build version 2 is released and packages have been rebuilt.
                raise DependencyException("Conda dependency failed to build job environment. "
                                          "This is most likely a limitation in conda. "
                                          "You can try to shorten the path to the job_working_directory.")
            raise DependencyException("Conda dependency seemingly installed but failed to build job environment.")

    def shell_commands(self, requirement):
        if not self.cache_path:
            # Build an isolated environment if not using a cached dependency manager
            self.build_environment()
        if self._preserve_python_environment:
            # On explicit testing the only such requirement I am aware of is samtools - and it seems to work
            # fine with just appending the PATH as done below. Other tools may require additional
            # variables in the future.
            return """export PATH=$PATH:'%s/bin' """ % (
                self.environment_path,
            )
        else:
            return """[ "$CONDA_DEFAULT_ENV" = "%s" ] || . %s '%s' > conda_activate.log 2>&1 """ % (
                self.environment_path,
                self.activate,
                self.environment_path
            )


def _string_as_bool( value ):
    return str( value ).lower() == "true"


__all__ = ('CondaDependencyResolver', 'DEFAULT_ENSURE_CHANNELS')



"""
Handle postfork functions under uWSGI
"""

# The uwsgi module is automatically injected by the parent uwsgi
# process and only exists that way.  If anything works, this is a
# uwsgi-managed process.
try:
    import uwsgi
    if hasattr(uwsgi, "numproc"):
        process_is_uwsgi = True
    else:
        process_is_uwsgi = False
except ImportError:
    # This is not a uwsgi process, or something went horribly wrong.
    process_is_uwsgi = False

try:
    from uwsgidecorators import postfork
except:
    def pf_dec(func):
        return func
    postfork = pf_dec
    if process_is_uwsgi:
        print("WARNING: This is a uwsgi process but the uwsgidecorators library"
              " is unavailable.  This is likely due to using an external (not"
              " in Galaxy's virtualenv) uwsgi and you may experience errors.")


postfork_functions = []


@postfork
def do_postfork():
    for f, args, kwargs in [ t for t in postfork_functions ]:
        f(*args, **kwargs)


def register_postfork_function(f, *args, **kwargs):
    if process_is_uwsgi:
        postfork_functions.append((f, args, kwargs))
    else:
        f(*args, **kwargs)



from __future__ import absolute_import

import collections
import copy
import json
import logging
import math
import random
import string

from six import iteritems, string_types, text_type

__all__ = ( "safe_dumps", "json_fix", "validate_jsonrpc_request", "validate_jsonrpc_response", "jsonrpc_request", "jsonrpc_response" )

log = logging.getLogger( __name__ )

to_json_string = json.dumps
from_json_string = json.loads


def json_fix( val ):
    if isinstance( val, list ):
        return [ json_fix( v ) for v in val ]
    elif isinstance( val, dict ):
        return dict( [ ( json_fix( k ), json_fix( v ) ) for ( k, v ) in iteritems(val) ] )
    elif isinstance( val, text_type ):
        return val.encode( "utf8" )
    else:
        return val


def swap_inf_nan( val ):
    """
    This takes an arbitrary object and preps it for jsonifying safely, templating Inf/NaN.
    """
    if isinstance(val, string_types):
        # basestring first, because it's a sequence and would otherwise get caught below.
        return val
    elif isinstance( val, collections.Sequence ):
        return [ swap_inf_nan( v ) for v in val ]
    elif isinstance( val, collections.Mapping ):
        return dict( [ ( swap_inf_nan( k ), swap_inf_nan( v ) ) for ( k, v ) in iteritems(val) ] )
    elif isinstance(val, float):
        if math.isnan(val):
            return "__NaN__"
        elif val == float("inf"):
            return "__Infinity__"
        elif val == float("-inf"):
            return "__-Infinity__"
        else:
            return val
    else:
        return val


def safe_loads( arg ):
    """
    This is a wrapper around loads that returns the parsed value instead of
    raising a value error. It also avoids autoconversion of non-iterables
    i.e numeric and boolean values.
    """
    try:
        loaded = json.loads( arg )
        if loaded is not None and not isinstance( loaded, collections.Iterable ):
            loaded = arg
    except ( TypeError, ValueError ):
        loaded = arg
    return loaded


def safe_dumps( *args, **kwargs ):
    """
    This is a wrapper around dumps that encodes Infinity and NaN values.  It's a
    fairly rare case (which will be low in request volume).  Basically, we tell
    json.dumps to blow up if it encounters Infinity/NaN, and we 'fix' it before
    re-encoding.
    """
    try:
        dumped = json.dumps( *args, allow_nan=False, **kwargs )
    except ValueError:
        obj = swap_inf_nan( copy.deepcopy( args[0] ) )
        dumped = json.dumps( obj, allow_nan=False, **kwargs )
    if kwargs.get( 'escape_closing_tags', True ):
        return dumped.replace( '</', '<\\/' )
    return dumped


# Methods for handling JSON-RPC

def validate_jsonrpc_request( request, regular_methods, notification_methods ):
    try:
        request = json.loads( request )
    except Exception as e:
        return False, request, jsonrpc_response( id=None,
                                                 error=dict( code=-32700,
                                                             message='Parse error',
                                                             data=str( e ) ) )
    try:
        assert 'jsonrpc' in request, \
            'This server requires JSON-RPC 2.0 and no "jsonrpc" member was sent with the Request object as per the JSON-RPC 2.0 Specification.'
        assert request['jsonrpc'] == '2.0', \
            'Requested JSON-RPC version "%s" != required version "2.0".' % request['jsonrpc']
        assert 'method' in request, 'No "method" member was sent with the Request object'
    except AssertionError as e:
        return False, request, jsonrpc_response( request=request,
                                                 error=dict( code=-32600,
                                                             message='Invalid Request',
                                                             data=str( e ) ) )
    try:
        assert request['method'] in ( regular_methods + notification_methods )
    except AssertionError as e:
        return False, request, jsonrpc_response( request=request,
                                                 error=dict( code=-32601,
                                                             message='Method not found',
                                                             data='Valid methods are: %s' % ', '.join( regular_methods + notification_methods ) ) )
    try:
        if request['method'] in regular_methods:
            assert 'id' in request, 'No "id" member was sent with the Request object and the requested method "%s" is not a notification method' % request['method']
    except AssertionError as e:
        return False, request, jsonrpc_response( request=request,
                                                 error=dict( code=-32600,
                                                             message='Invalid Request',
                                                             data=str( e ) ) )
    return True, request, None


def validate_jsonrpc_response( response, id=None ):
    try:
        response = json.loads( response )
    except Exception as e:
        log.error( 'Response was not valid JSON: %s' % str( e ) )
        log.debug( 'Response was: %s' % response )
        return False, response
    try:
        assert 'jsonrpc' in response, \
            'This server requires JSON-RPC 2.0 and no "jsonrpc" member was sent with the Response object as per the JSON-RPC 2.0 Specification.'
        assert ( 'result' in response or 'error' in response ), \
            'Neither of "result" or "error" members were sent with the Response object.'
        if 'error' in response:
            assert int( response['error']['code'] ), \
                'The "code" member of the "error" object in the Response is missing or not an integer.'
            assert 'message' in response, \
                'The "message" member of the "error" object in the Response is missing.'
    except Exception as e:
        log.error( 'Response was not valid JSON-RPC: %s' % str( e ) )
        log.debug( 'Response was: %s' % response )
        return False, response
    if id is not None:
        try:
            assert 'id' in response and response['id'] == id
        except Exception as e:
            log.error( 'The response id "%s" does not match the request id "%s"' % ( response['id'], id ) )
            return False, response
    return True, response


def jsonrpc_request( method, params=None, id=None, jsonrpc='2.0' ):
    if method is None:
        log.error( 'jsonrpc_request(): "method" parameter cannot be None' )
        return None
    request = dict( jsonrpc=jsonrpc, method=method )
    if params:
        request['params'] = params
    if id is not None and id is True:
        request['id'] = ''.join( [ random.choice( string.hexdigits ) for i in range( 16 ) ] )
    elif id is not None:
        request['id'] = id
    return request


def jsonrpc_response( request=None, id=None, result=None, error=None, jsonrpc='2.0' ):
    if result:
        rval = dict( jsonrpc=jsonrpc, result=result )
    elif error:
        rval = dict( jsonrpc=jsonrpc, error=error )
    else:
        msg = 'jsonrpc_response() called with out a "result" or "error" parameter'
        log.error( msg )
        rval = dict( jsonrpc=jsonrpc, error=msg )
    if id is not None:
        rval['id'] = id
    elif request is not None and 'id' in request:
        rval['id'] = request['id']
    return rval



"""Compat. layer with conda_build/verify if Galaxy/galaxy-lib not installed through conda.

In general there are utilities available for Conda building and parsing that are high-quality
and should be utilized when available but that are only available in conda channels and not in
PyPI. This module serves as a PyPI capable interface to these utilities.
"""
import collections
import os

import yaml

try:
    from conda_build.metadata import MetaData
except ImportError:
    MetaData = None

try:
    from anaconda_verify.recipe import parse, render_jinja2
except ImportError:
    render_jinja2 = None
    parse = None


class _Memoized(object):

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value


def _parse(data, cfg):
    """Parse metadata YAML."""
    assert cfg is None, "Conda utilities for evaluating cfg are not available."
    return dict(yamlize(data))


def _render_jinja2(recipe_dir):
    """Evaluate Conda recipe as a jinja template."""
    try:
        import jinja2
    except ImportError:
        raise Exception("Failed to import jinja2 for evaluating Conda recipe templates.")

    loaders = [jinja2.FileSystemLoader(recipe_dir)]
    env = jinja2.Environment(loader=jinja2.ChoiceLoader(loaders))
    template = env.get_or_select_template('meta.yaml')
    return template.render(environment=env)


@_Memoized
def yamlize(data):
    res = yaml.load(data)
    # ensure the result is a dict
    if res is None:
        res = {}
    return res


if render_jinja2 is None:
    render_jinja2 = _render_jinja2

if parse is None:
    parse = _parse


def raw_metadata(recipe_dir):
    """Evaluate Conda template if needed and return raw metadata for supplied recipe directory."""
    meta_path = os.path.join(recipe_dir, 'meta.yaml')
    with open(meta_path, 'rb') as fi:
        data = fi.read()
        if b'{{' in data:
            data = render_jinja2(recipe_dir)
    meta = parse(data, None)
    return meta


class _MetaData(object):

    def __init__(self, input_dir):
        self.meta = raw_metadata(input_dir)

    def get_value(self, field, default=None):
        """Get nested field value or supplied default is not present."""
        section, key = field.split('/')
        submeta = self.meta.get(section)
        if submeta is None:
            submeta = {}
        res = submeta.get(key)
        if res is None:
            res = default
        return res


if MetaData is None:
    MetaData = _MetaData

__all__ = (
    "MetaData",
    "raw_metadata",
)



import json
import os
import shutil

from .cwltool_deps import ref_resolver
from .parser import (
    JOB_JSON_FILE,
    load_job_proxy,
)


def handle_outputs(job_directory=None):
    # Relocate dynamically collected files to pre-determined locations
    # registered with ToolOutput objects via from_work_dir handling.
    if job_directory is None:
        job_directory = os.path.join(os.getcwd(), os.path.pardir)
    cwl_job_file = os.path.join(job_directory, JOB_JSON_FILE)
    if not os.path.exists(cwl_job_file):
        # Not a CWL job, just continue
        return
    # So we only need to do strict validation when the tool was loaded,
    # no reason to do it again during job execution - so this shortcut
    # allows us to not need Galaxy's full configuration on job nodes.
    job_proxy = load_job_proxy(job_directory, strict_cwl_validation=False)
    tool_working_directory = os.path.join(job_directory, "working")
    outputs = job_proxy.collect_outputs(tool_working_directory)
    for output_name, output in outputs.items():
        target_path = job_proxy.output_path(output_name)
        if isinstance(output, dict) and "location" in output:
            output_path = ref_resolver.uri_file_path(output["location"])
            if output["class"] != "File":
                open("galaxy.json", "w").write(json.dump({
                    "dataset_id": job_proxy.output_id(output_name),
                    "type": "dataset",
                    "ext": "expression.json",
                }))
            shutil.move(output_path, target_path)
            for secondary_file in output.get("secondaryFiles", []):
                # TODO: handle nested files...
                secondary_file_path = ref_resolver.uri_file_path(secondary_file["location"])
                assert secondary_file_path.startswith(output_path)
                secondary_file_name = secondary_file_path[len(output_path):]
                secondary_files_dir = job_proxy.output_secondary_files_dir(
                    output_name, create=True
                )
                extra_target = os.path.join(secondary_files_dir, secondary_file_name)
                shutil.move(
                    secondary_file_path,
                    extra_target,
                )
        else:
            with open(target_path, "w") as f:
                f.write(json.dumps(output))


__all__ = (
    'handle_outputs',
)



from galaxy.tools.deps.requirements import ToolRequirements
from galaxy.util import bunch


class DependenciesDescription(object):
    """ Capture (in a readily serializable way) context related a tool
    dependencies - both the tool's listed requirements and the tool shed
    related context required to resolve dependencies via the
    ToolShedPackageDependencyResolver.

    This is meant to enable remote resolution of dependencies, by the Pulsar or
    other potential remote execution mechanisms.
    """

    def __init__(self, requirements=[], installed_tool_dependencies=[]):
        self.requirements = requirements
        # tool shed installed tool dependencies...
        self.installed_tool_dependencies = installed_tool_dependencies

    def to_dict(self):
        return dict(
            requirements=[r.to_dict() for r in self.requirements],
            installed_tool_dependencies=[DependenciesDescription._toolshed_install_dependency_to_dict(d) for d in self.installed_tool_dependencies]
        )

    @staticmethod
    def from_dict(as_dict):
        if as_dict is None:
            return None

        requirements_dicts = as_dict.get('requirements', [])
        requirements = ToolRequirements.from_list(requirements_dicts)
        installed_tool_dependencies_dicts = as_dict.get('installed_tool_dependencies', [])
        installed_tool_dependencies = map(DependenciesDescription._toolshed_install_dependency_from_dict, installed_tool_dependencies_dicts)
        return DependenciesDescription(
            requirements=requirements,
            installed_tool_dependencies=installed_tool_dependencies
        )

    @staticmethod
    def _toolshed_install_dependency_from_dict(as_dict):
        # Rather than requiring full models in Pulsar, just use simple objects
        # containing only properties and associations used to resolve
        # dependencies for tool execution.
        repository_object = bunch.Bunch(
            name=as_dict['repository_name'],
            owner=as_dict['repository_owner'],
            installed_changeset_revision=as_dict['repository_installed_changeset'],
        )
        dependency_object = bunch.Bunch(
            name=as_dict['dependency_name'],
            version=as_dict['dependency_version'],
            type=as_dict['dependency_type'],
            tool_shed_repository=repository_object,
        )
        return dependency_object

    @staticmethod
    def _toolshed_install_dependency_to_dict(tool_dependency):
        tool_shed_repository = tool_dependency.tool_shed_repository
        return dict(
            dependency_name=tool_dependency.name,
            dependency_version=tool_dependency.version,
            dependency_type=tool_dependency.type,
            repository_name=tool_shed_repository.name,
            repository_owner=tool_shed_repository.owner,
            repository_installed_changeset=tool_shed_repository.installed_changeset_revision,
        )
